"""
This module provides storage functionality for the Common-EGSE.

![Storage Manager](../../../img/storage-design.png)

### Introduction

All control servers and other components that need to save data through the storage manager
need to register to the storage manager first. That can be done by creating a `StorageProxy`
and sending a register message for the component. The name that is registered is also used by
all `save` messages to identify the component and dispatch the correct persistence function.

All control servers (that inherit from the ControlServer class) will automatically register to
the Storage Manager when the control server (CS) starts. When the CS quits normally (by the
`quit_server()` command of the Service proxy) the control server will automatically be
unregistered from the Storage manager.

By default, when the Storage is active, it will save all the information that it gets for those
components that have registered to the Storage manager.

```python
storage_proxy = StorageProxy()
...
storage_proxy.register({'origin': reg_name,
                        'persistence_class': TYPES['CSV'],
                        'prep': {'column_names': [], 'mode': 'w'}})
...

storage_proxy.save({'origin': reg_name, 'data': data})
...
storage_proxy.unregister({'origin': reg_name})
```

### Commands

The Storage component shall understand the following commands:

* registration of a component

    * `storage_proxy.register({'origin': str, 'persistence_class'; <class>, 'prep': dict})`
    * `storage_proxy.unregister({'origin': str})`

* accept the following commands from the configuration manager

    * `storage_proxy.start_observation(obsid: str)`
    * `storage_proxy.end_observation(obsid: str)`

* accept housekeeping and data packets from the DPU simulator housekeeping and accept
    housekeeping and status data from any control server

    * `storage_proxy.save({'origin': name, 'data': data})`


When an observation is started, the Storage Manager will 'fork' the data stream and save the
control server information in a separate set of files. The filenames will carry the `obsid`
(see below).

### What data is saved?

* Housekeeping data from other control servers, e.g. device CS or the configuration manager
* Status data from other control servers
* SpaceWire packets from the N-FEE and the F-FEE, this includes housekeeping and CCD data
* Image data from the camera, i.e. processed Spacewire data packets

### How is data saved?

* Different data types are stored in specific formats according to the PersistenceLayer
  that was chosen
* The following file formats are supported:
    * CSV - for tabular data, typically housekeeping and status information
    * TXT - for logging information
    * FITS - for image data
    * HDF5 - for SpaceWire data and housekeeping packets

### How are the files named?

We have two sets of files:

1. files that contain only the data that was collected for an observation, i.e.
   between the calls to `start_observation` and `end_observation`. These files are located
   in the `obs` sub-folder of the main data store location. The filename is constructed from
   the test id, the site id and the setup id, followed by the data source identifier and a
   timestamp. An example from the PUNA Hexapod housekeeping file:
   `00031_CSL_00008_PUNA_20200701_210711.csv`.

2. files that contain all the housekeeping for each of the data sources regardless of an
   observation is running or not. All data that is collected during a test day will be stored in
   these files. Outside an observation context there will be no CCD image data collected.
   These files are located in the `daily` sub-folder of the main data store location. The filename
   is constructed from the date, the site id and the data source identifier. An example for the
   same PUNA Hexapod file: `20200701_CSL_PUNA.csv`.

The timestamp that is used is the time of file creation.

"""
from __future__ import annotations

__all__ = [
    "is_storage_manager_active",
    "get_status",
    "PersistenceLayer",
    "store_housekeeping_information",
    "register_to_storage_manager",
    "unregister_from_storage_manager",
]

import abc
import datetime
import logging
import os
import shutil
import textwrap
from functools import partial
from pathlib import Path
from pathlib import PurePath
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

from egse.bits import humanize_bytes
from egse.command import ClientServerCommand
from egse.config import find_files
from egse.control import ControlServer
from egse.env import get_site_id
from egse.listener import Event
from egse.listener import EventInterface
from egse.response import Failure
from egse.response import Response
from egse.response import Success
from egse.control import is_control_server_active
from egse.decorators import dynamic_interface
from egse.env import get_data_storage_location
from egse.exceptions import Error
from egse.obsid import ObservationIdentifier
from egse.obsid import TEST_LAB
from egse.persistence import PersistenceLayer
from egse.protocol import CommandProtocol
from egse.proxy import Proxy
from egse.proxy import REQUEST_TIMEOUT
from egse.settings import Settings
from egse.setup import Setup
from egse.setup import get_setup
from egse.storage.persistence import TYPES
from egse.system import format_datetime
from egse.zmq_ser import bind_address
from egse.zmq_ser import connect_address

logger = logging.getLogger(__name__)

HERE = Path(__file__).parent

CTRL_SETTINGS = Settings.load("Storage Control Server")
SITE_ID = get_site_id()
DEVICE_SETTINGS = COMMAND_SETTINGS = Settings.load(location=HERE, filename="storage.yaml")


def is_storage_manager_active(timeout: float = 0.5):
    """Check if the Storage Manager is running.

    Returns:
        True if the Storage Manager is running and replied with the expected answer.
    """

    endpoint = connect_address(
        CTRL_SETTINGS.PROTOCOL, CTRL_SETTINGS.HOSTNAME, CTRL_SETTINGS.COMMANDING_PORT
    )

    return is_control_server_active(endpoint, timeout)


def register_to_storage_manager(origin: str, persistence_class: PersistenceLayer, prep: Dict):
    """
    Register the component to the Storage manager.

    For information on what should go into the `prep` keyword argument, please check the proper
    persistence class.

    Args:
        origin (str): the name of the component, by which it will be registered
        persistence_class: a concrete class that will be used to store the  data
        prep (dict): preparation meta data for the persistence class

    Raises:
        ConnectionError: when the storage manager can not be reached.
    """

    try:
        with StorageProxy() as proxy:
            rc = proxy.register(
                {
                    "origin": origin,
                    "persistence_class": persistence_class,
                    "prep": prep,
                }
            )
            if not rc.successful:
                logger.warning(f"Couldn't register to the Storage manager: {rc}")
            else:
                logger.info(rc)
    except ConnectionError as exc:
        logger.warning(f"Couldn't connect to the Storage manager for registration: {exc}")
        raise


def unregister_from_storage_manager(origin: str):
    """
    Unregister the component from the Storage manager.

    Args:
        origin (str): the name of the component, by which it will be registered

    Raises:
        ConnectionError: when the storage manager can not be reached.
    """

    try:
        with StorageProxy() as proxy:
            rc = proxy.unregister({"origin": origin})
            if not rc.successful:
                logger.warning(f"Couldn't unregister from the Storage manager: {rc}")
            else:
                logger.info(rc)
    except ConnectionError as exc:
        logger.warning(f"Couldn't connect to the Storage manager for de-registration: {exc}")
        raise


def store_housekeeping_information(origin: str, data: dict):
    """
    Send housekeeping information to the Storage manager. The housekeeping data is usually collected by the device
    control server and is a dictionary with key:value = parameters names and their values. The data dictionary shall
    at least contain a timestamp in the format generated by `format_datetime()`.

    Raises:
        ConnectionError: when the Storage manager can not be reached.
    """

    # logger.debug("Sending housekeeping data to storage manager.")

    try:
        with StorageProxy() as proxy:
            rc = proxy.save({"origin": origin, "data": data})
            if not rc.successful:
                logger.warning(f"Couldn't save data to the Storage manager for {origin=}, cause: {rc}")
    except ConnectionError as exc:
        logger.warning(f"Couldn't connect to the Storage manager to store housekeeping: {exc}")
        raise


def cycle_daily_files():
    """
    Create a new daily file for each registered item when no such file exists.
    """
    with StorageProxy() as storage:
        storage.cycle_daily_files()


class AlreadyRegisteredError(Error):
    """This error indicates that an item is already registered and cannot be registered again."""


class Registry:
    """
    A registry for registration of components in the system that need to save data through
    the Storage Manager.
    """

    def __init__(self):
        self._register = dict()

    def __contains__(self, name: str):
        """Returns True if an item with 'name' has been registered."""
        if isinstance(name, str):
            return name in self._register.keys()

        raise ValueError(
            f"You can only check if something is contained in the Registry "
            f"by a key of type string, item is of type '{type(name)}'."
        )

    def __len__(self):
        """Returns the number of registrations."""
        return len(self._register)

    def __iter__(self):
        return iter(self._register.keys())

    def get(self, name: str):
        """Returns the registered item for the given name (identifier)."""
        return self._register.get(name)

    def register(self, name: str, item):
        """Register the item by the given name in the register.

        Args:
            name (str): the key to identify this registration, usually the name of
                the control server or the class that registers
            item: an object that contains information about what information needs to be saved
                and how
        """
        if not isinstance(name, str):
            raise ValueError("The name of the item to register must be a string.")
        if name in self:
            raise AlreadyRegisteredError(
                f"An item with name '{name}' is already registered, please unregister first."
            )
        self._register[name] = item

    def unregister(self, name: str):
        """Unregister the item with the given name from the register.

        Args:
            name (str): the key by which the registration was done.
        """
        if not isinstance(name, str):
            raise ValueError("The name of the item to unregister must be a string.")
        if name not in self:
            raise KeyError(f"There is no item with name '{name}' in this Register.")
        del self._register[name]


class StoragePacket(metaclass=abc.ABCMeta):
    """Base packet for all data send to Storage."""

    def __init__(self, origin=None, data=None, metadata=None):
        self._timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self._origin = origin
        self._data = data
        self._metadata = metadata

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def origin(self):
        return self._origin

    @property
    def data(self):
        return self._data

    @property
    def metadata(self):
        return self._metadata


class StorageInterface:
    """
    This interface is for control servers to register to the Storage Manager and for the
    configuration manager to start and stop an observation/test.

    The interface should be implemented by the StorageController and the StorageProxy (and
    possibly a StorageSimulator should we need that).
    """

    @dynamic_interface
    def save(self, item: dict) -> Response:
        """Saves the data part from the item.
        Args:
            item (dict): a dictionary that identifies the origin and the data to be stored.
        """
        raise NotImplementedError

    @dynamic_interface
    def read(self, item: dict) -> Response:
        """Reads data from storage defined by the `origin` in item.
        Args:
            item (dict): a dictionary that identifies the origin and an optional filter.
        """
        raise NotImplementedError

    @dynamic_interface
    def register(self, item: dict, use_counter: bool = False) -> Response:
        """Registers the item to the storage manager.

        The item shall have the following keys:

        * origin (str): the name of the item, by which it will be registered
        * persistence_class (class):

        Args:
            item (dict): a dictionary that identifies the component to be registered

        Returns:
            Success: when the item could be registered successfully
            Failure: when the registration fails, `Failure.cause` provides cause exception
        """
        raise NotImplementedError

    @dynamic_interface
    def unregister(self, item: dict) -> Response:
        """Unregisters the item from the storage manager.

        Args:
            item (dict): a dictionary that identifies the component to be registered

        Returns:
            Success: when the item could be unregistered successfully
            Failure: when the de-registration fails, `Failure.cause` provides cause exception
        """
        raise NotImplementedError

    @dynamic_interface
    def get_registry_names(self):
        """Returns the names of the registered components.

        Returns:
            a list of names/identifiers for the registered components.
        """
        raise NotImplementedError

    @dynamic_interface
    def start_observation(self, obsid: ObservationIdentifier, camera_name: str = None) -> Response:
        """ "Start an observation for the given obsid.

        When a new obsevation is started the following actions will be taken:

        * Housekeeping and telemetry from registered components (mainly control servers)
          will be forked into a newly created file for that component.
        * Camera data will be saved as follows:
            * SpaceWire packets will go into an HDF5 file for this observation
            * CCD data will be assembled into image data and saved in FITS format

        Args:
            camera_name: the name of the camera or None
            obsid: a unique observation identifier

        Returns:
            Success: when a new observation with the given obsid could be started properly.
            Failure: when the previous observation was not finished yet (no end_observation was
                send), or when a failure happened during the preparations for a new observation.
        """
        raise NotImplementedError

    @dynamic_interface
    def end_observation(self, obsid: ObservationIdentifier) -> Response:
        """Ends the currently running observation.

        When a running observation is ended, the following actions will be taken:

        * All files of registered components will be closed.
        * Housekeeping and telemetry will continue to be saved to the global repository

        Args:
            obsid: the observation identifier of the currently running observation

        Returns:
            Success: when the current observation could be ended successfully.
            Failure: when the given obsid doesn't match the current observation.

        """
        raise NotImplementedError

    @dynamic_interface
    def get_obsid(self):
        """Return the observation identifier."""
        pass

    @dynamic_interface
    def cycle_daily_files(self):
        pass

    @dynamic_interface
    def get_storage_location(self):
        pass

    @dynamic_interface
    def get_filenames(self, item: dict) -> List[Path]:
        """Return the filename(s) associated with this registered item."""
        pass

    @dynamic_interface
    def new_registration(self, item: dict, use_counter=False):
        """
        Create a new data file for the given item. If the item was previously registered, close that
        registration and open a new registration. The use_counter parameter determines if an
        incremented counter is used to construct a unique filename.

        Args:
            - item: Dictionary that identifies the component to be registered.
            - use_counter: Indicates whether or not a counter should be used in the filename.
        """

        pass

    @dynamic_interface
    def get_disk_usage(self):
        """ Return the total, used, and free disk space [bytes].

        Returns:
            - Total disk space [bytes].
            - Used disk space [bytes].
            - Free disk space [bytes].
        """

        pass

    @dynamic_interface
    def get_loaded_setup_id(self) -> str:
        """
        Returns the ID of the currently loaded Setup.

        Note:
            This is the Setup active on this control server. This command is mainly used to check that the Setup
            loaded in this control server corresponds to the Setup loaded in the configuration manager.

        Returns:
            The ID of the Setup loaded in this control server.
        """

        pass


def _disentangle_filename(filename: Union[str, Path]) -> Tuple:
    """Disentangle the given filename and return the test identifier, the site id and the Setup id.

    It is assumed in this function that the filename is from a test observation and contains the
    correct fields to be extracted. Only very limited checking is done if that is indeed the case.

    Args:
        filename (str, Path): the filename of a test observation

    Returns:
        A tuple containing the test_id (int), site_id (str) and setup_id (int). If the filename is
        not recognized as a valid filename, the returned tuple contains all None.

    """
    filename = Path(filename).resolve()
    parts = filename.parts

    if parts[-2] != 'obs':
        name = parts[-1]
        if not (name.rsplit('_', 3)[0].endswith('_SPW') or name.rsplit('_', 2)[0].endswith('_SPW')):
            return None, None, None

    name = parts[-1]
    test_id, site_id, setup_id = name.split('_')[:3]
    return int(test_id), site_id, int(setup_id)


def _construct_filename(
        identifier: str, ext: str, obsid: ObservationIdentifier = None, use_counter=False,
        location: str = None, site_id: str = None, camera_name: str = None
) -> PurePath:
    """Construct a filename for the data source.

    We construct two types of filenames:

    1. the observational files which store all the data that are collected during an observation.
       There is one file per data source. The files are located in the `obs` sub-folder of the
       storage location.
    2. the daily files which store all the data from a data source regardless the state of an
       observation. There is one file per data source. The files are located in the `daily`
       sub-folder of the storage location.

    Args:
        identifier (str): an identifier for the source of the data, this string is usually what
            is sent in the `origin` of the item dictionary.
        ext (str): the extension of the file, this depends oon the persistence class that is
            used for storing the data.
        obsid (ObservationIdentifier): a unique identifier for the observation
        use_counter: Indicates whether or not a counter should be included in the filename.
    Returns:
        The full path to the file as a `PurePath`.
    """

    site_id = site_id or SITE_ID
    location = location or get_data_storage_location(site_id=site_id)

    if obsid:

        timestamp = datetime.datetime.now(tz=datetime.timezone.utc).strftime("%Y%m%d_%H%M%S")

        prefix = obsid.create_id(order=TEST_LAB, camera_name=camera_name)
        location = location / Path("obs") / f"{prefix}"
        if not os.path.exists(location):
            os.makedirs(location)

        if use_counter:
            counter_file_path = location / f"{prefix}_{identifier}.count"
            if not counter_file_path.exists():
                pattern = f"{prefix}_{identifier}_{timestamp}_*.{ext}"
                counter = determine_counter_from_dir_list(location, pattern)
                _write_counter(counter, counter_file_path)
            else:
                counter = get_counter(counter_file_path)
            name = f"{prefix}_{identifier}_{counter:05d}_{timestamp}.{ext}"
        else:
            name = f"{prefix}_{identifier}_{timestamp}.{ext}"

    else:

        timestamp = datetime.datetime.now(tz=datetime.timezone.utc).strftime("%Y%m%d")

        location = location / Path("daily") / timestamp
        if not os.path.exists(location):
            os.makedirs(location)

        if use_counter:
            counter_file_path = location / f"{timestamp}_{site_id}_{identifier}.count"
            if not counter_file_path.exists():
                pattern = f"{timestamp}_{site_id}_{identifier}_*.{ext}"
                counter = determine_counter_from_dir_list(location, pattern)
                _write_counter(counter, counter_file_path)
            else:
                counter = get_counter(counter_file_path)
            name = f"{timestamp}_{site_id}_{identifier}_{counter:05d}.{ext}"
        else:
            name = f"{timestamp}_{site_id}_{identifier}.{ext}"

    return Path(location) / name


def _write_counter(counter: int, file_path: Path):
    """
    Overwrites the given counter in the given file. The file contains nothing else then the counter.
    If the file didn't exist before, it will be created.

    Args:
        counter: the counter to save
        file_path: the file to which the counter shall be saved
    """
    with file_path.open('w') as fd:
        fd.write(f"{counter:d}")


def _read_counter(file_path: Path) -> int:
    """
    Reads a counter from the given file. The file shall only contain the counter which must
    be an integer on the first line of the file. If the given file doesn't exist, 0 is returned.

    Args:
        file_path: the full path of the file containing the counter

    Returns:
        The counter that is read from the file or 0 if file doesn't exist.
    """
    try:
        with file_path.open('r') as fd:
            counter = fd.read().strip()
    except FileNotFoundError:
        counter = 0
    return int(counter or 0)


def get_counter(file_path: Path) -> int:
    """
    Read the counter from a dedicated file, add one and save the counter back to the file..

    Args:
        - file_path: full pathname of the file that contains the required counter

    Returns:
        The value of the next counter, 1 if no previous files were found or if an error occurred.
    """

    counter = _read_counter(file_path)
    counter += 1
    _write_counter(counter, file_path)

    return counter


def determine_counter_from_dir_list(location, pattern, index: int = -1):
    """
    Determine counter for a new file at the given location and with the given pattern.
    The next counter is determined from the sorted list of files that match the given pattern.

    Args:
        - location: Location where the file should be stored.
        - pattern: Pattern for the filename.
        - index: the location of the counter in the filename after it is split on '_' [default=-1]

    Returns:
        The value of the next counter, 1 if no previous files were found or if an error occurred.
    """

    files = sorted(find_files(pattern=pattern, root=location))

    # No filenames found showing the given pattern -> start counting at 1

    if len(files) == 0:
        return 1

    last_file = files[-1]

    parts = last_file.name.split("_")

    try:

        # Observation files have the following pattern:
        #  <test ID>_<lab ID>_<setup ID>_<storage mnemonic>_<day YYYYmmdd>_<time HHMMSS>[_<counter>]
        # Daily files:
        #  <day>_<site ID>_<storage mnemonic>[_<counter>]

        counter = int(parts[index].split(".")[0]) + 1
        logger.debug(f"{counter = }")
        return counter

    except ValueError:
        logger.warning("ValueError", exc_info=True)
        return 1


class StorageController(StorageInterface, EventInterface):
    """
    The Storage Controller handles the registration of components, the start and end of an
    observation/test and the dispatching of the persistence functions in save.
    """

    def __init__(self, control_server):
        self._obsid: ObservationIdentifier | None = None
        self._camera_name: str | None = None
        self._registry = Registry()
        self._cs: ControlServer = control_server
        self._setup: Setup | None = None

    def start_observation(self, obsid: ObservationIdentifier, camera_name: str = None) -> Response:
        if self._obsid is not None:
            return Failure(
                "Can not start a new observation before the previous observation is ended."
            )

        self._obsid = obsid
        self._camera_name = camera_name
        if camera_name != self._setup.camera.ID.lower():
            logger.error(
                f"Mismatch in camera name between Setup in Storage Manager {self._setup.camera.ID.lower()} "
                f"and Setup in Configuration Manager {camera_name}!"
            )

        # open a dedicated file for each registered item

        for registered_name in self._registry:
            registered_item = self._registry.get(registered_name)

            if "persistence_count" in registered_item:
                # no need to fork any files that contain persistence_counts
                continue

            # FIXME: create a list for all classes that need to be skipped here. Allow the list to be changed
            #        by a method call
            no_duplicates = tuple()
            if issubclass(registered_item["persistence_class"], no_duplicates):
                # do not duplicate HDF5 files during an observation - issue #1186
                continue

            # NOTE: The following lines of code contain tests for special treatment of HDF5 files
            # while the above check disables HDF5 files in OBS. We leave the code in for now until
            # a definite decision is taken.

            filename = _construct_filename(
                registered_item["origin"],
                registered_item["persistence_class"].extension,
                obsid,
                use_counter=issubclass(registered_item["persistence_class"], TYPES["HDF5"]),
                camera_name=camera_name
            )

            # logger.debug(f"{filename = }, {camera_name = }")

            # we have more than one file open for this item DAILY and OBS.... take care of that

            # Special case for HDF5 files as they need to be copied instead of created

            if issubclass(registered_item["persistence_class"], TYPES["HDF5"]):
                daily_file_object: PersistenceLayer = registered_item["persistence_objects"][0]
                daily_file_path: Path = daily_file_object.get_filepath()
                logger.debug(f"Copying {daily_file_path} to {filename}")

                # Close the HDF5 file before copy, otherwise you will get
                # a 'bad object header version number' when opening the destination.

                daily_file_object.close()
                shutil.copy(daily_file_path, filename)
                daily_file_object.open(mode='a')

            persistence_obj = registered_item["persistence_class"](
                filename, prep=registered_item["prep"]
            )

            mode = "a" if persistence_obj.exists() else "w"
            persistence_obj.open(mode=mode)

            registered_item["persistence_objects"].append(persistence_obj)

        return Success("Storage successfully started observation.")

    def end_observation(self, obsid: ObservationIdentifier) -> Response:
        if obsid != self._obsid:
            return Failure(f"Given obsid doesn't match current obsid: {obsid} != {self._obsid}")

        # close the dedicated file for each registered item

        for registered_name in self._registry:
            registered_item = self._registry.get(registered_name)
            if "persistence_count" in registered_item:
                # no need to close any files that contain persistence_counts
                continue
            try:
                persistence_obj = registered_item["persistence_objects"].pop()
                persistence_obj.close()
            except IndexError as exc:
                logger.warning(f"Trying to close a persistent object for {registered_name}, {exc=}")

        self._obsid = None
        self._camera_name = None

        return Success("Storage successfully ended observation.")

    def get_obsid(self):
        return self._obsid

    def save(self, item: dict) -> Response:
        """Saves the data contained in this item to the right location and format.

        Args:
            item: dictionary with at least the following keywords - origin, data
        Returns:
            Success: when the data has been properly saved.
        """
        # TODO:
        #  this method might become a performance problem and the reason that we might have
        #  back pressure problem. Keep an eye on this and do performance tests.

        # What needs to be done:
        # * based on item['origin'], check if item component is registered
        # * get register entry for item
        # * for persistence in persistence list:
        #      persistence.create(data)

        registered_item = self._registry.get(item["origin"])

        if not registered_item:
            return Failure(
                f"Storage could not find a registration for {item['origin']}, no data saved."
            )

        for persistence_object in registered_item["persistence_objects"]:
            persistence_object.create(item["data"])

        return Success(f"Storage successfully saved the data for {item['origin']}.")

    def read(self, item: dict):

        registered_item = self._registry.get(item["origin"])

        if not registered_item:
            return Failure(
                f"Storage could not find a registration for {item['origin']}, no data saved."
            )

        # FIXME:
        #   * wat als meerdere persistence_objects bestaan? alleen de eerste, alleen de laatste,
        #     samenvoegen? een nieuw keyword in item?

        result = None
        for persistence_object in registered_item["persistence_objects"]:
            result = persistence_object.read(item["select"])

        return Success(f"Storage successfully read the data from {item['origin']}.", result)

    def register(self, item: dict, use_counter=False) -> Response:

        if not isinstance(item, dict):
            return Failure(
                f"Could not register item, item must be a dictionary (item={type(item)})."
            )

        prep = item.get("prep", {})

        # When we register an item, the file should always be 'created', unless this is a
        # persistence count and we just need to append to the file, always.

        # if "persistence_count" not in item:
        #     prep.update({"mode": "w"})

        if "origin" not in item or "persistence_class" not in item:
            return Failure("Could not register item, missing mandatory keyword(s).")

        try:
            self._registry.register(item["origin"], item)

            if "filename" in item:
                location = Path(get_data_storage_location(site_id=SITE_ID))
                filename = location / item["filename"]
            else:
                filename = _construct_filename(item["origin"], item["persistence_class"].extension,
                                               use_counter=use_counter)

            persistence_obj = item["persistence_class"](filename, prep=prep)
            mode = "a" if persistence_obj.exists() else "w"
            persistence_obj.open(mode=mode)

            # add the PersistenceLayer object to the registered item

            item["persistence_objects"] = [persistence_obj]

            # Special case when the components registration is done after an observation was
            # started, unless we are handling a persistence_count, in which case there is only one
            # file.

            if (
                self._obsid
                and "persistence_count" not in item
                and not issubclass(item["persistence_class"], TYPES["HDF5"])
            ):
                filename = _construct_filename(
                    item["origin"], item["persistence_class"].extension, self._obsid,
                    use_counter=use_counter, camera_name=self._camera_name
                )

                persistence_obj = item["persistence_class"](filename, prep=prep)
                mode = "a" if persistence_obj.exists() else "w"
                persistence_obj.open(mode=mode)

                item["persistence_objects"].append(persistence_obj)

            msg = f"Storage successfully registered {item['origin']}"
            logger.info(msg)
            return Success(msg)
        except AlreadyRegisteredError as exc:
            msg = f"Could not register {item['origin']}: {exc}"
            logger.error(msg)
            return Success(f"{item['origin']} is already registered.")
        except (ValueError, KeyError) as exc:
            # FIXME:
            #  Should I unregister here? and if yes, should I not call the
            #  self.unregister(item) method instead?
            self._registry.unregister(item["origin"])
            msg = f"Could not register {item['origin']}: {exc}"
            logger.error(msg)
            return Failure(f"Could not register {item['origin']}", exc)

    def unregister(self, item) -> Response:
        try:
            registered_item = self._registry.get(item["origin"])
            if registered_item is None:
                raise ValueError("The item is not registered.")

            # Probably also should close the file and some other things

            for persistence_obj in registered_item.get("persistence_objects", []):
                persistence_obj.close()

            self._registry.unregister(item["origin"])

            msg = f"Storage successfully unregistered {item['origin']}"
            logger.info(msg)
            return Success(msg)
        except (ValueError, KeyError) as exc:
            return Failure(f"Could not unregister {item['origin']}", exc)

    def get_registry_names(self):
        return list(self._registry)

    def cycle_daily_files(self):

        logger.info("Cycling daily files for Storage Manager")

        for reg_name in self._registry:
            item = self._registry.get(reg_name)
            if "persistence_count" in item:
                # no need to cycle any files that contain persistence_counts
                continue
            if "persistence_objects" in item:
                logger.info(f"Cycling daily file for {item['origin']}.")

                # The first item in the list is always the daily persistence object, however, for
                # the N-FEE_SPW origin, sometimes when the N-FEE is not ON, there is no persistence
                # object. (see issue #1458) So, we catch this and continue.

                try:
                    daily_persist_obj = item["persistence_objects"][0]
                    daily_persist_obj.close()
                except IndexError:
                    logger.info(f"I'm ignoring that there is no persistence_object "
                                f"for {item['origin']} at this time.")
                    continue

                # Create folder for the day
                filename = _construct_filename(item['origin'], item['persistence_class'].extension)

                persistence_obj: PersistenceLayer = item["persistence_class"](
                    filename, prep=item.get("prep")
                )
                mode = "a" if persistence_obj.exists() else "w"
                persistence_obj.open(mode=mode)

                # replace the previous daily persistence object with the current

                item["persistence_objects"][0] = persistence_obj

            else:
                # We should never get here, since when an item is registered, the 'file' is
                # opened and it should exist
                logger.error(
                    f"Found a registered item {item} that has no persistence objects.",
                    stack_info=True,
                )

    def get_storage_location(self):
        return get_data_storage_location(site_id=SITE_ID)

    def get_filenames(self, item: dict) -> List[Path]:
        registered_item = self._registry.get(item["origin"])

        if not registered_item:
            return []

        return [
            persistence_object.get_filepath()
            for persistence_object in registered_item["persistence_objects"]
        ]

    def new_registration(self, item: dict, use_counter=False) -> Response:
        if item["origin"] in self.get_registry_names():
            _ = self.unregister(item)

        response = self.register(item, use_counter=use_counter)
        logger.info(f"From register: {response=}")
        return response

    def get_disk_usage(self):

        location = Path(get_data_storage_location(site_id=SITE_ID))
        total, used, free = shutil.disk_usage(location)
        return total, used, free

    def get_loaded_setup_id(self) -> str:
        return self._setup.get_id() if self._setup is not None else "no setup loaded"

    def load_setup(self, setup_id: int = 0):
        # Use get_setup() here instead of load_setup() in order to prevent recursively notifying and loading Setups.
        # That is because the load_setup() method will notify the listeners that a new Setup has been loaded.
        try:
            setup = get_setup()
        except Exception as exc:
            raise RuntimeError(f"Exception caught: {exc!r}")

        if setup is None:
            raise RuntimeError("Couldn't get Setup from the configuration manager.")

        if isinstance(setup, Failure):
            raise setup

        # time.sleep(20.0)  # used as a test to check if this method is blocking the commanding... it is!

        # logger.info(f"{setup_id = }, {setup.get_id() = }")

        if 0 < setup_id != int(setup.get_id()):
            raise RuntimeError(f"Setup IDs do not match: {setup.get_id()} != {setup_id}, no Setup loaded.")
        else:
            self._setup = setup
            logger.info(f"Setup {setup.get_id()} loaded in the Storage manager.")

    def handle_event(self, event: Event) -> str:
        logger.info(f"An event is received, {event=}")
        try:
            if event.type == 'new_setup':
                self._cs.schedule_task(partial(self.load_setup, setup_id=event.context['setup_id']))
        except KeyError as exc:
            return f"Expected event context to contain the following key: {exc}"
        return "ACK"


class StorageCommand(ClientServerCommand):
    pass


class StorageProxy(Proxy, StorageInterface, EventInterface):
    """The StorageProxy class is used to connect to the Storage Manager (control server) and
    send commands remotely."""

    def __init__(
        self,
        protocol=CTRL_SETTINGS.PROTOCOL,
        hostname=CTRL_SETTINGS.HOSTNAME,
        port=CTRL_SETTINGS.COMMANDING_PORT,
        timeout=REQUEST_TIMEOUT,
    ):
        """
        Args:
            protocol: the transport protocol [default is taken from settings file]
            hostname: location of the control server (IP address)
                [default is taken from settings file]
            port: TCP port on which the control server is listening for commands
                [default is taken from settings file]
        """
        super().__init__(connect_address(protocol, hostname, port), timeout=timeout)


class StorageProtocol(CommandProtocol):
    def __init__(self, control_server: ControlServer):
        super().__init__()
        self.control_server = control_server

        self.controller = StorageController(control_server)

        self.load_commands(COMMAND_SETTINGS.Commands, StorageCommand, StorageController)

        self.build_device_method_lookup_table(self.controller)

    def get_bind_address(self):
        return bind_address(
            self.control_server.get_communication_protocol(),
            self.control_server.get_commanding_port(),
        )

    def get_status(self) -> dict:
        return super().get_status()

    def get_housekeeping(self) -> dict:
        return {
            "timestamp": format_datetime(),
        }


def get_status(full: bool = False):

    if is_storage_manager_active():
        with StorageProxy() as sm:
            text =  textwrap.dedent(
                f"""\
                Storage Manager:
                    Status: [green]active[/]
                    Hostname: {sm.get_ip_address()}
                    Monitoring port: {sm.get_monitoring_port()}
                    Commanding port: {sm.get_commanding_port()}
                    Service port: {sm.get_service_port()}
                    Storage location: {sm.get_storage_location()}
                    Loaded Setup: {sm.get_loaded_setup_id()}
                    Registrations: {sm.get_registry_names()}
                """
            )
            if full:
                text += "Filenames for all registered items:\n"
                for origin in sm.get_registry_names():
                    fn = [x for x in sm.get_filenames({"origin": origin})]
                    text += f"  {origin:10s} ->  {fn}\n"
                if obsid := sm.get_obsid():
                    text += f"An observation is registered: {obsid}\n"
                else:
                    text += "No observation is registered.\n"

                total, used, free = sm.get_disk_usage()
                text += f"Total disk space: {humanize_bytes(total)}\n"
                text += f"Used disk space: {humanize_bytes(used)} ({(used / total * 100):.2f}%)\n"
                text += f"Free disk space: {humanize_bytes(free)} ({(free / total * 100):.2f}%)\n"

        return text

    else:
        return "Storage Manager Status: [red]not active"



"""
The Control Server that connects to the Hexapod PUNA Hardware Controller.

Start the control server from the terminal as follows:

    $ puna_cs start-bg

or when you don't have the device available, start the control server in simulator mode. That
will make the control server connect to a device software simulator:

    $ puna_cs start --sim

Please note that software simulators are intended for simple test purposes and will not simulate
all device behavior correctly, e.g. timing, error conditions, etc.

"""
import logging
from typing import Annotated

import typer

from egse.hexapod.symetrie import ProxyFactory
from egse.hexapod.symetrie import get_hexapod_controller_pars

if __name__ != "__main__":
    import multiprocessing

    multiprocessing.current_process().name = "puna_cs"

import sys

import rich
import zmq

from egse.control import is_control_server_active
from egse.zmq_ser import connect_address

from prometheus_client import start_http_server

from egse.control import ControlServer
from egse.hexapod.symetrie.puna_protocol import PunaProtocol
from egse.settings import Settings

_LOGGER = logging.getLogger(__name__)

CTRL_SETTINGS = Settings.load("Hexapod PUNA Control Server")


class PunaControlServer(ControlServer):
    """
    PunaControlServer - Command and monitor the Hexapod PUNA hardware.

    This class works as a command and monitoring server to control the Symétrie Hexapod PUNA.
    This control server shall be used as the single point access for controlling the hardware
    device. Monitoring access should be done preferably through this control server also,
    but can be done with a direct connection through the PunaController if needed.

    The sever binds to the following ZeroMQ sockets:

    * a REQ-REP socket that can be used as a command server. Any client can connect and
      send a command to the Hexapod.

    * a PUB-SUP socket that serves as a monitoring server. It will send out Hexapod status
      information to all the connected clients every five seconds.

    """

    def __init__(self, simulator: bool = False):
        super().__init__()

        self.device_protocol = PunaProtocol(self, simulator=simulator)

        self.logger.info(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

    def get_communication_protocol(self):
        return CTRL_SETTINGS.PROTOCOL

    def get_commanding_port(self):
        return CTRL_SETTINGS.COMMANDING_PORT

    def get_service_port(self):
        return CTRL_SETTINGS.SERVICE_PORT

    def get_monitoring_port(self):
        return CTRL_SETTINGS.MONITORING_PORT

    def get_storage_mnemonic(self):
        try:
            return CTRL_SETTINGS.STORAGE_MNEMONIC
        except AttributeError:
            return "PUNA"

    def before_serve(self):
        start_http_server(CTRL_SETTINGS.METRICS_PORT)


app = typer.Typer()


@app.command()
def start(
        simulator: Annotated[
            bool,
            typer.Option("--simulator", "--sim", help="start the hexapod PUNA simulator in the background")
        ] = False
):
    """
    Start the Hexapod PUNA Control Server.

    Args:
        simulator: start the hexapod PUNA simulator in the background.
    """

    try:

        controller = PunaControlServer(simulator)
        controller.serve()

    except KeyboardInterrupt:

        print("Shutdown requested...exiting")

    except SystemExit as exit_code:

        print("System Exit with code {}.".format(exit_code))
        sys.exit(exit_code)

    except Exception:

        _LOGGER.exception("Cannot start the Hexapod Puna Control Server")

        # The above line does exactly the same as the traceback, but on the _LOGGER
        # import traceback
        # traceback.print_exc(file=sys.stdout)

    return 0


@app.command()
def stop():
    """Send a 'quit_server' command to the Hexapod Puna Control Server."""

    *_, device_id, device_name, controller_type = get_hexapod_controller_pars()

    factory = ProxyFactory()

    try:
        with factory.create(device_name, device_id=device_id) as proxy:
            sp = proxy.get_service_proxy()
            sp.quit_server()
    except ConnectionError:
        rich.print("[red]Couldn't connect to 'puna_cs', process probably not running. ")


@app.command()
def status():
    """Request status information from the Control Server."""

    protocol = CTRL_SETTINGS.PROTOCOL
    hostname_cs = CTRL_SETTINGS.HOSTNAME
    port_cs = CTRL_SETTINGS.COMMANDING_PORT

    endpoint = connect_address(protocol, hostname_cs, port_cs)

    *_, device_id, device_name, controller_type = get_hexapod_controller_pars()

    factory = ProxyFactory()

    if is_control_server_active(endpoint):
        rich.print("PUNA Hexapod: [green]active")
        with factory.create(device_name, device_id=device_id) as puna:
            sim = puna.is_simulator()
            connected = puna.is_connected()
            ip = puna.get_ip_address()
            rich.print(f"type: {controller_type}")
            rich.print(f"mode: {'simulator' if sim else 'device'}{'' if connected else ' not'} connected")
            rich.print(f"hostname: {ip}")
            rich.print(f"commanding port: {port_cs}")
    else:
        rich.print("PUNA Hexapod: [red]not active")


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG, format=Settings.LOG_FORMAT_FULL)

    sys.exit(app())



import datetime
import logging
import re
import socket

import contextlib

import rich
import typer

from egse.settings import Settings
from egse.system import SignalCatcher

logging.basicConfig(level=logging.DEBUG)

_LOGGER = logging.getLogger("egse.daq6510.sim")

HOST = "localhost"
DAQ_SETTINGS = Settings.load("Keithley DAQ6510")


device_time = datetime.datetime.now(datetime.timezone.utc)
reference_time = device_time


app = typer.Typer(help="DAQ6510 Simulator")


def create_datetime(year, month, day, hour, minute, second):
    global device_time, reference_time
    device_time = datetime.datetime(year, month, day, hour, minute, second, tzinfo=datetime.timezone.utc)
    reference_time = datetime.datetime.now(datetime.timezone.utc)


def nothing():
    return None


def set_time(year, month, day, hour, minute, second):
    print(f"TIME {year}, {month}, {day}, {hour}, {minute}, {second}")
    create_datetime(int(year), int(month), int(day), int(hour), int(minute), int(second))


def get_time():
    current_device_time = device_time + (datetime.datetime.now(datetime.timezone.utc) - reference_time)
    msg = current_device_time.strftime("%a %b %d %H:%M:%S %Y")
    print(f":SYST:TIME? {msg = }")
    return msg


def beep(a, b):
    print(f"BEEP {a=}, {b=}")


def reset():
    print("RESET")


COMMAND_ACTIONS_RESPONSES = {
    "*IDN?": (None, "KEITHLEY INSTRUMENTS, MODEL DAQ6510, SIMULATOR"),
}


COMMAND_PATTERNS_ACTIONS_RESPONSES = {
    r":?\*RST": (reset, None),
    r":?SYST(?:em)*:TIME (\d+), (\d+), (\d+), (\d+), (\d+), (\d+)": (set_time, None),
    r":?SYST(?:em)*:TIME\? 1": (nothing, get_time),
    r":?SYST(?:em)*:BEEP(?:er)* (\d+), (\d+(?:\.\d+)?)": (beep, None),
}


def write(conn, response: str):

    response = f"{response}\n".encode()
    _LOGGER.debug(f"write: {response = }")
    conn.sendall(response)


def read(conn) -> str:
    """
    Reads one command string from the socket, i.e. until a linefeed ('\n') is received.

    Returns:
        The command string with the linefeed stripped off.
    """

    idx, n_total = 0, 0
    buf_size = 1024 * 4
    command_string = bytes()

    try:
        for _ in range(100):
            data = conn.recv(buf_size)
            n = len(data)
            n_total += n
            command_string += data
            # if data.endswith(b'\n'):
            if n < buf_size:
                break
    except socket.timeout as e_timeout:
        # LOGGER.warning(f"Socket timeout error from {e_timeout}")
        # This timeout is catched at the caller, where the timeout is set.
        raise

    # _LOGGER.debug(f"Total number of bytes received is {n_total}, idx={idx}")
    _LOGGER.info(f"read: {command_string=}")

    return command_string.decode().rstrip()


def process_command(command_string: str) -> str:

    global COMMAND_ACTIONS_RESPONSES
    global COMMAND_PATTERNS_ACTIONS_RESPONSES

    # LOGGER.debug(f"{command_string=}")

    try:
        action, response = COMMAND_ACTIONS_RESPONSES[command_string]
        action and action()
        if error_msg:
            return error_msg
        else:
            return response if isinstance(response, str) else response()
    except KeyError:
        # try to match with a value
        for key, value in COMMAND_PATTERNS_ACTIONS_RESPONSES.items():
            if match := re.match(key, command_string):
                # LOGGER.debug(f"{match=}, {match.groups()}")
                action, response = value
                # LOGGER.debug(f"{action=}, {response=}")
                action and action(*match.groups())
                return error_msg or (response if isinstance(response, str) or response is None else response())
        return f"ERROR: unknown command string: {command_string}"


def run_simulator():
    global error_msg

    _LOGGER.info("Starting the DAQ6510 Simulator")

    killer = SignalCatcher()

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind((HOST, DAQ_SETTINGS.PORT))
        s.listen()
        s.settimeout(2.0)
        while True:
            while True:
                with contextlib.suppress(socket.timeout):
                    conn, addr = s.accept()
                    break
                if killer.term_signal_received:
                    return
            with conn:
                _LOGGER.info(f'Accepted connection from {addr}')
                write(conn, 'This is PLATO DAQ6510 X.X.sim')
                conn.settimeout(2.0)
                try:
                    error_msg = ""
                    while True:
                        with contextlib.suppress(socket.timeout):
                            data = read(conn)
                            _LOGGER.info(f"{data = }")
                            if data.strip() == "STOP":
                                _LOGGER.info("Client requested to terminate...")
                                s.close()
                                return
                            for cmd in data.split(';'):
                                response = process_command(cmd.strip())
                                if response is not None:
                                    write(conn, response)
                            if not data:
                                _LOGGER.info("Client closed connection, accepting new connection...")
                                break
                        if killer.term_signal_received:
                            _LOGGER.info("Terminating...")
                            s.close()
                            return
                        if killer.user_signal_received:
                            if killer.signal_name == "SIGUSR1":
                                _LOGGER.info("SIGUSR1 is not supported by this simulator")
                            if killer.signal_name == "SIGUSR2":
                                _LOGGER.info("SIGUSR2 is not supported by this simulator")
                            killer.clear()

                except ConnectionResetError as exc:
                    _LOGGER.info(f"ConnectionResetError: {exc}")
                except Exception as exc:
                    _LOGGER.info(f"{exc.__class__.__name__} caught: {exc.args}")


def send_request(cmd: str):

    from egse.tempcontrol.keithley.daq6510_devif import DAQ6510EthernetInterface

    daq_dev = DAQ6510EthernetInterface(hostname="localhost", port=5025)
    daq_dev.connect()

    response = daq_dev.query(cmd)

    daq_dev.disconnect()

    return response


@app.command()
def start():
    run_simulator()


@app.command()
def status():
    response = send_request("*IDN?")
    print(f"{response}")


@app.command()
def stop():
    response = send_request("STOP")
    print(f"{response}")


if __name__ == '__main__':
    app()



from __future__ import annotations

import contextlib
import logging
import pickle
import queue
import random
import threading
import time

import zmq

from egse.system import format_datetime
from egse.zmq_ser import MessageIdentifier

_LOGGER = logging.getLogger(__name__)


def _g_tick(period: float):
    """Generator for ticks every period [s]."""
    next_time = time.monotonic()
    while True:
        next_time += period
        yield max(next_time - time.monotonic(), 0)


class HeartbeatBroadcaster(threading.Thread):
    """
    Sends a heartbeat signal to the endpoint using a PUB-SUB protocol.

    The heartbeat message is sent by default every second, but that can be changed
    by the `period` argument (which is in fractional seconds).

    Since the broadcaster runs in a thread, it is not a full prove method that the parent process is still running.
    Therefore, the parent process can set a custom messages on the queue. That custom message will be broadcast at
    the same time as the next heartbeat, which is probably not the time that the custom message was set. Keep that
    in mind.
    """

    def __init__(self, period: float = 1.0, endpoint: str = "tcp://*:5555"):
        super().__init__()
        self._period = period
        self._endpoint = endpoint
        self._socket: zmq.Socket | None = None
        self._running = False
        self._canceled = threading.Event()
        self._queue = queue.Queue()
        self._heartbeat_id = MessageIdentifier.HEARTBEAT.to_bytes(1, byteorder='big')
        self._custom_id = MessageIdentifier.CUSTOM.to_bytes(1, byteorder='big')

    def run(self):
        ctx = zmq.Context.instance()
        self._socket = ctx.socket(zmq.PUB)
        self._socket.bind(self._endpoint)
        self._running = True

        g = _g_tick(self._period)
        while True:
            if self._canceled.is_set():
                break
            time.sleep(next(g))
            self.send_heartbeat()
            self.send_custom_message()

    def send_heartbeat(self):
        timestamp = format_datetime()
        self._socket.send_multipart([self._heartbeat_id, pickle.dumps(f"Heartbeat – {timestamp}")])

    def send_custom_message(self):
        with contextlib.suppress(queue.Empty):
            message = self._queue.get_nowait()
            self._socket.send_multipart([self._custom_id, pickle.dumps(message)])

    def set_message(self, message: str):
        self._queue.put_nowait(message)

    def cancel(self) -> None:
        self._canceled.set()


if __name__ == '__main__':

    broadcaster = HeartbeatBroadcaster()
    broadcaster.start()

    # You will see that the timestamp of this custom message is drifting because
    # we do not take the execution time of the `set_message()` etc. intyo accout.
    # In the actual HeartbeatBroadcaster above, we do take that into account.

    try:
        while True:
            broadcaster.set_message(f"Custom message – {format_datetime()}")
            time.sleep(1.0 * random.choice([1, 2, 3, 4]))
    except KeyboardInterrupt:
        broadcaster.cancel()



"""
This module provides dummy implementation for classes of the Commanding chain.
"""
import logging
import random
import sys

import click
import zmq

from egse.command import ClientServerCommand
from egse.control import ControlServer
from egse.control import is_control_server_active
from egse.decorators import dynamic_interface
from egse.protocol import CommandProtocol
from egse.proxy import Proxy
from egse.settings import Settings
from egse.system import AttributeDict
from egse.system import format_datetime
from egse.zmq_ser import bind_address
from egse.zmq_ser import connect_address

logging.basicConfig(level=logging.DEBUG, format=Settings.LOG_FORMAT_FULL)
LOGGER = logging.getLogger("egse.dummy")

# Especially DummyCommand and DummyController need to be defined in a known module
# because those objects are pickled and when de-pickled at the clients side the class
# definition must be known.

# We use AttributeDict here to define the settings, because that is how the Settings.load() returns
# settings loaded from a YAML file.

ctrl_settings = AttributeDict(
    {
        'HOSTNAME': 'localhost',
        'COMMANDING_PORT': 4443,
        'SERVICE_PORT': 4444,
        'MONITORING_PORT': 4445,
        'PROTOCOL': 'tcp',
        'TIMEOUT': 10,
        'HK_DELAY': 1.0,
    }
)

commands = AttributeDict(
    {
        'info': {
            'description': 'Info on the Dummy Controller',
            'response': 'handle_device_method'
        },
        'response': {
            'description': 'send a command to the device and return it\'s response',
            'device_method': 'response',
            'cmd': '{one} {two} {fake}',
            'response': 'handle_device_method'
        }
    }
)


def is_dummy_cs_active():
    return is_control_server_active(
        endpoint=connect_address(ctrl_settings.PROTOCOL, ctrl_settings.HOSTNAME, ctrl_settings.COMMANDING_PORT)
    )


class DummyCommand(ClientServerCommand):
    pass


class DummyInterface:
    @dynamic_interface
    def info(self):
        ...
    @dynamic_interface
    def response(self, *args, **kwargs):
        ...


class DummyProxy(Proxy, DummyInterface):
    def __init__(self,
                 protocol=ctrl_settings.PROTOCOL, hostname=ctrl_settings.HOSTNAME, port=ctrl_settings.COMMANDING_PORT):
        """
        Args:
            protocol: the transport protocol [default is taken from settings file]
            hostname: location of the control server (IP address) [default is taken from settings file]
            port: TCP port on which the control server is listening for commands [default is taken from settings file]
        """
        super().__init__(connect_address(protocol, hostname, port), timeout=ctrl_settings.TIMEOUT)


class DummyController(DummyInterface):
    def info(self):
        return "method info() called on DummyController"

    def response(self, *args, **kwargs):
        return f"response({args}, {kwargs})"


class DummyProtocol(CommandProtocol):

    def __init__(self, control_server: ControlServer):
        super().__init__()
        self.control_server = control_server

        self.device_controller = DummyController()

        self.load_commands(commands, DummyCommand, DummyController)

        self.build_device_method_lookup_table(self.device_controller)

        self._count = 0

    def get_bind_address(self):
        return bind_address(self.control_server.get_communication_protocol(), self.control_server.get_commanding_port())

    def get_status(self):
        return super().get_status()

    def get_housekeeping(self) -> dict:

        LOGGER.debug(f"Executing get_housekeeping function for {self.__class__.__name__}.")

        self._count += 1

        # use the sleep to test the responsiveness of the control server when even this get_housekeeping function takes
        # a lot of time, i.e. up to several minutes in the case of data acquisition devices
        # import time
        # time.sleep(2.0)


        return {
            'timestamp': format_datetime(),
            'COUNT': self._count,
            'PI': 3.14159,  # just to have a constant parameter
            'Random': random.randint(0, 100),  # just to have a variable parameter
        }


class DummyControlServer(ControlServer):
    """
    DummyControlServer - Command and monitor dummy device controllers.

    The sever binds to the following ZeroMQ sockets:

    * a REQ-REP socket that can be used as a command server. Any client can connect and
      send a command to the dummy device controller.

    * a PUB-SUP socket that serves as a monitoring server. It will send out status
      information to all the connected clients every DELAY seconds.

    """

    def __init__(self):
        super().__init__()

        self.device_protocol = DummyProtocol(self)

        self.logger.info(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.set_hk_delay(ctrl_settings.HK_DELAY)

    def get_communication_protocol(self):
        return 'tcp'

    def get_commanding_port(self):
        return ctrl_settings.COMMANDING_PORT

    def get_service_port(self):
        return ctrl_settings.SERVICE_PORT

    def get_monitoring_port(self):
        return ctrl_settings.MONITORING_PORT

    def get_storage_mnemonic(self):
        return "DUMMY-HK"


@click.group()
def cli():
    pass


@cli.group()
def control_server():
    pass


@control_server.command()
def start():
    """Start the dummy control server on localhost."""

    try:
        control_server = DummyControlServer()
        control_server.serve()
    except KeyboardInterrupt:
        print("Shutdown requested...exiting")
    except SystemExit as exit_code:
        print(f"System Exit with code {exit_code}.")
        sys.exit(1)
    except Exception:
        import traceback
        traceback.print_exc(file=sys.stdout)


@control_server.command()
def stop():
    """Send a quit service command to the dummy control server."""
    with DummyProxy() as dummy:
        sp = dummy.get_service_proxy()
        sp.quit_server()


if __name__ == "__main__":
    cli()



import shutil
from pathlib import Path

import pytest

from egse.config import WorkingDirectory
from egse.config import find_dirs
from egse.config import find_file
from egse.config import find_files
from egse.config import find_first_occurrence_of_dir
from egse.config import find_root
from egse.config import get_common_egse_root

_HERE = Path(__file__).parent.resolve()


def test_find_first_occurrence_of_dir():

    assert str(find_first_occurrence_of_dir("conf", root=_HERE)).endswith("tests/data/conf")
    assert str(find_first_occurrence_of_dir("dev1", root=_HERE)).endswith("tests/data/lib/dev1")
    assert str(find_first_occurrence_of_dir("dev2", root=_HERE)).endswith("tests/data/lib/dev2")

    assert find_first_occurrence_of_dir("not-a-directory", root=_HERE) is None

    # Pass in a different root directory

    assert str(find_first_occurrence_of_dir("dev1", root=_HERE/"data")).endswith("lib/dev1")

    folders = (
        _HERE / "x_data/01/kul",
        _HERE / "x_data/02/kul",
        _HERE / "x_data/02/42/kul",
        _HERE / "x_data/03/42/kul",
        _HERE / "x_data/03/43/kul",
        _HERE / "x_data/03/43/kul/ivs",
        _HERE / "x_data/04/42/kal",
        _HERE / "x_data/04/42/kul",
    )
    for folder in folders:
        folder.mkdir(parents=True)

    assert str(find_first_occurrence_of_dir("kul", root=_HERE)).endswith("tests/x_data/01/kul")
    assert str(find_first_occurrence_of_dir("03/42/kul", root=_HERE)).endswith("tests/x_data/03/42/kul")
    assert str(find_first_occurrence_of_dir("03/*/kul", root=_HERE)).endswith("tests/x_data/03/42/kul")
    assert str(find_first_occurrence_of_dir("42/kul", root=_HERE)).endswith("tests/x_data/02/42/kul")
    assert str(find_first_occurrence_of_dir("*/42/kul", root=_HERE)).endswith("tests/x_data/02/42/kul")
    assert str(find_first_occurrence_of_dir("ivs", root=_HERE)).endswith("tests/x_data/03/43/kul/ivs")
    assert str(find_first_occurrence_of_dir("04/*/k?l", root=_HERE)).endswith("tests/x_data/04/42/kal")

    shutil.rmtree(_HERE / "x_data")


def test_find_root():

    assert find_root(None) is None
    assert find_root("/") is None
    assert find_root("/", tests=("tmp",)) == Path("/")
    assert find_root("/", tests=("non-existing-tmp",)) is None


def test_get_common_egse_root():

    print()

    # for the following test I assume that we are in the repository, but we can not test
    # the value as it will be different for each installation
    assert get_common_egse_root() is not None

    print(f"{get_common_egse_root() = }")

    # for the following test I assume that we are on a unix system (Linux or macOS)
    assert get_common_egse_root(Path("/tmp")) is None


def test_find_root_exceptions():

    assert find_root("/non-existing-path") is None
    assert find_root(None) is None


def test_get_common_egse_root_with_env():

    import os

    os.environ["COMMON_EGSE_PATH"] = "/Users/rik/git"

    # I added lru_cache to speed up the get_common_egse_root() function, but this
    # is of course fatal for test harnesses. T_HEREfore, clear the cache before and
    # after this test.

    get_common_egse_root.cache_clear()

    assert get_common_egse_root() == Path("/Users/rik/git")

    get_common_egse_root.cache_clear()

    del os.environ["COMMON_EGSE_PATH"]


def test_find_files():
    files = list(find_files("COPY*", root=get_common_egse_root()))
    assert files

    for f in files:
        assert f.name.startswith("COPY")

    # no files named 'data', only folders that are named 'data', use find_dirs for this.

    files = list(find_files("data", root=get_common_egse_root() / "src"))
    assert not files

    # When I want to find a file in a specific directory, use the in_dir keyword

    filename_pattern = "shared-lib.so"
    files = list(find_files(filename_pattern, in_dir="lib/dev1"))
    print()
    print(files)

    # The expected file is in the src/egse/lib/CentOS-7 folder, but
    # t_HERE could also be a build directory which contains the file.

    assert len(files) in (1, 2)


def test_find_dirs():
    dir_name = "dev[12]"
    dirs = list(find_dirs(dir_name))
    assert dirs

    dir_name = "dev1"
    dirs = list(find_dirs(dir_name))
    assert dirs

    dir_name = "lib/dev*"
    dirs = list(find_dirs(dir_name))
    print(dirs)
    # The third file could be in the build folder which doesn't always exists.
    # A fourth file could be in the virtual environment venv or venv38
    assert len(dirs) in (2, 3, 4)

    # use the leading '/' to prevent that another 'lib/dev' is matched.

    dir_name = "/lib/dev*"
    dirs = list(find_dirs(dir_name))
    print(dirs)
    # The second file could be in the build folder which doesn't always exists.
    assert len(dirs) in (1, 2)


def test_find_file():

    assert find_file("pyproject.toml")
    assert find_file("data-file.txt")
    assert not find_file("non-existing-file.txt")

    assert find_file("config.py", in_dir="egse")


def test_working_directory():
    import os

    # Check if a ValueError is raised when using a non-existing folder

    with pytest.raises(ValueError):
        with WorkingDirectory("/XXX"):  # There shall not be such a directory at the root folder
            pass

    # Check if indeed the context manager has changed directories

    cwd = os.getcwd()

    with WorkingDirectory(_HERE.parent) as wdir:
        assert wdir.path / "tests" == _HERE
        for file in wdir.path.glob('tests'):
            assert str(file) == str(_HERE)

    assert cwd == os.getcwd()



import pytest

from egse.listener import EVENT_ID
from egse.listener import Event
from egse.listener import Listeners


def test_listeners():

    listeners = Listeners()
    assert len(listeners) == 0

    # Success stories ----------------------------------------------------------

    listeners.add_listener({'name': 'test_process_1'})
    assert len(listeners) == 1

    assert "test_process_1" in listeners.get_listener_names()

    listeners.remove_listener({'name': 'test_process_1'})
    assert len(listeners) == 0

    assert "test_process_1" not in listeners.get_listener_names()

    listeners.add_listener({'name': 'test_process_2'})
    listeners.add_listener({'name': 'test_process_3'})

    assert "test_process_2" in listeners.get_listener_names()
    assert "test_process_3" in listeners.get_listener_names()

    # A KeyError will be raised because no Proxy was defined
    with pytest.raises(KeyError):
        listeners.notify_listeners(Event(EVENT_ID.ALL, "General notification"))

    # A KeyError will be raised because no Proxy was defined
    with pytest.raises(KeyError):
        listeners.notify_listeners(Event(event_id=EVENT_ID.SETUP, context={"new_item": "This is a new item."}))

    # Exceptions ---------------------------------------------------------------

    listeners.add_listener({'name': 'test_process_1'})
    with pytest.raises(ValueError):
        listeners.add_listener({'name': 'test_process_1'})

    with pytest.raises(ValueError):
        listeners.add_listener({'name': 'test_process_2'})

    listeners.remove_listener({'name': 'test_process_1'})
    with pytest.raises(ValueError):
        listeners.remove_listener({'name': 'test_process_1'})

    with pytest.raises(ValueError):
        listeners.add_listener({})

    with pytest.raises(ValueError):
        listeners.add_listener({'name': 'invalid_proxy', 'proxy': 'should be a Proxy class'})



import asyncio
# import subprocess
import sys

import rich


async def run_all_status(full: bool = False):
    tasks = [
        status_log_cs(),
        status_sm_cs(full),
        status_cm_cs(),
    ]

    await asyncio.gather(*tasks)


async def status_log_cs():

    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'egse.logger.log_cs', 'status',
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    # proc = subprocess.Popen(
    #     [sys.executable, '-m', 'egse.logger.log_cs', 'status'],
    #     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
    # )
    #
    # stdout, stderr = proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_sm_cs(full: bool = False):

    cmd = [sys.executable, '-m', 'egse.storage.storage_cs', 'status']
    if full:
        cmd.append("--full" if full else "")

    proc = await asyncio.create_subprocess_exec(
        *cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    # proc = subprocess.Popen(
    #     [sys.executable, '-m', 'egse.storage.storage_cs', 'status'],
    #     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
    # )
    #
    # stdout, stderr = proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_cm_cs():

    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'egse.confman.confman_cs', 'status',
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    # proc = subprocess.Popen(
    #     [sys.executable, '-m', 'egse.confman.confman_cs', 'status'],
    #     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
    # )
    #
    # stdout, stderr = proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr:
        rich.print(f"[red]{stderr.decode()}[/]")



"""
This module defines Mixin classes that can be used for adding methods and properties to
classes without strict inheritance.

Warning:
    Be careful, some of the Mixin classes require certain attributes to be defined in the
    outer subclass. Read the docstrings carefully to understand what is needed.
"""

import contextlib
import enum
import functools
import inspect
import logging
import string
from typing import Callable
from typing import Dict

from egse.command import ClientServerCommand
from egse.command import CommandError
from egse.command import CommandExecution
from egse.protocol import DynamicCommandProtocol
from egse.protocol import get_function

LOGGER = logging.getLogger(__name__)

__all__ = [
    "CommandType",
    "DynamicCommandMixin",
    "add_cr_lf",
    "add_eot",
    "add_etx",
    "add_lf",
    "dynamic_command",
]

# ----- Mixin for dynamic commanding ---------------------------------------------------------------

COMMAND_TYPES = {
    "read": "__read_command",
    "write": "__write_command",
    "query": "__query_command",
    "transaction": "__transaction_command",
}
STX = "\x02"  # start-of-text
ETX = "\x03"  # end-of-text
EOT = "\x04"  # end-of-transmission
LINE_FEED = "\x0A"
CARRIAGE_RETURN = "\x0D"


def add_etx(cmd_string: str) -> str:
    """
    Add an end-of-text ETX (ASCII code 0x03) to the command string.

    Args:
        cmd_string: the unprocessed command string

    Returns:
        The command string with the ETX character appended.
    """
    return cmd_string if cmd_string.endswith(ETX) else cmd_string + ETX


def add_eot(cmd_string: str) -> str:
    """
    Add an end-of-transmission EOT (ASCII code 0x04) to the command string.

    Args:
        cmd_string: the unprocessed command string

    Returns:
        The command string with the EOT character appended.
    """
    return cmd_string if cmd_string.endswith(EOT) else cmd_string + EOT


def add_lf(cmd_string: str) -> str:
    """ Add a line feed to the given command string, if not added yet.

    Args:
        cmd_string: Command string.

    Returns:
        Given command string with a line feed appended (if not present yet).
    """

    if not cmd_string.endswith(LINE_FEED):

        cmd_string += LINE_FEED

    return cmd_string


def add_cr_lf(cmd_string: str) -> str:
    """ Add a carriage return and line feed to the given command string, if not added yet.

    Args:
        cmd_string: Command string.

    Returns:
        Given command string with a carriage return and line feed appended (if not present yet).
    """

    if not cmd_string.endswith(CARRIAGE_RETURN + LINE_FEED):

        cmd_string += CARRIAGE_RETURN + LINE_FEED

    return cmd_string


def expand_kwargs(kwargs: Dict):
    """Expand keyword arguments and their values as 'key=value' separated by spaces."""
    return " ".join(f"{k}={v}" for k, v in kwargs.items())


class CommandType(str, enum.Enum):
    """The type of the command, i.e. read, write, or transaction."""
    READ = "read"
    WRITE = "write"
    TRANSACTION = "transaction"


def dynamic_command(
        *,
        cmd_type: str,  # required keyword-only argument
        cmd_string: str = None,
        process_response: Callable = None,
        process_cmd_string: Callable = None,
        process_kwargs: Callable = None,
        use_format: bool = False,
        pre_cmd: Callable = None,
        post_cmd: Callable = None,
):
    """Convert an interface method into a dynamic command.

    The arguments define the type of command and how the response shall be processed.

    The command types 'write', 'query', and 'transaction' must be accompanied by a `cmd_string`
    argument that defines the formatting of the eventual command string that will be passed to
    the transport functions. The `cmd_string` is a template string that contains `$`-based
    substitutions for the function arguments. When you specify the `use_format=True` keyword,
    the `cmd_string` will be formatted using the format() function instead of the template
    substitution. The format option is less secure, but provides the functionality to format
    the arguments.

    A template string looks like:

        cmd_string="CREATE:SENS:TEMP ${name} ${type} default=${default}"

    The same `cmd_string` as a format option:

        cmd_string="CREATE:SENS:TEMP {name} {type} default={default:0.4f}"
        use_format=True

    The process_response and process_cmd_string keywords allow you to specify a pure function to
    process the response before it is returned, and to process the cmd_string before it is sent
    to the transport function.

    The `pre_cmd` and `post_cmd` keywords specify a callable/function to be executed before and/or
    after the actual command was executed (i.e. send to the device). These functions are called
    with specific keyword arguments that allow additional device interaction and response processing.
    The `pre_cmd` function is called with the keyword argument `transport=` which passes the device
    transport. This allows the function to interact with the device again through the methods defined
    by the DeviceTransport interface. The `pre_cmd` function must not return anything.
    The `post_cmd` function is called with the keyword arguments `transport=` and `response=`.
    The response argument contains the response from the command that was previously sent to the
    device. The `post_cmd` function can use this response to parse its content and act against
    this content, although possible, it is usually not a good idea to alter the content of the
    response argument. The `post_cmd` function shall return (i.e. pass through) the response.

    This decorator can add the following static attributes to the method:

    * `__dynamic_interface`
    * `__read_command`, `__write_command`, `__query_command`, `__transaction_command`
    * `__cmd_string`
    * `__process_response`
    * `__process_cmd_string`
    * `__use_format`
    * `__pre_cmd`
    * `__post_cmd`

    Args:
        cmd_type (str): one of 'read', 'write', 'query', or 'transaction' [required keyword]
        cmd_string (str): format string for the generation of the instrument command
        process_response (Callable): function to process the response
        process_cmd_string (Callable): function to process the command string after substitution
        process_kwargs (Callable): function to expand the kwargs after substitution
        use_format (bool): use string formatting instead of string templates
        pre_cmd (Callable): the function to execute before the command is executed
        post_cmd (Callable): the function to execute after the command is executed
    """

    # Perform some checks on required arguments

    if cmd_type not in COMMAND_TYPES:
        raise ValueError(f"Unknown type passed into dynamic command decorator: {type=}")

    if cmd_type in ('write', 'query', 'transaction') and cmd_string is None:
        raise ValueError(f"No cmd_string was provided for {cmd_type=}.")

    def func_wrapper(func: Callable):
        """Adds the different static attributes."""

        setattr(func, "__dynamic_interface", True)

        setattr(func, COMMAND_TYPES[cmd_type], True)

        if cmd_string is not None:
            setattr(func, "__cmd_string", cmd_string)

        if process_response is not None:
            setattr(func, "__process_response", process_response)

        if process_cmd_string is not None:
            setattr(func, "__process_cmd_string", process_cmd_string)

        if process_kwargs is not None:
            setattr(func, "__process_kwargs", process_kwargs)

        if use_format:
            setattr(func, "__use_format", True)

        if pre_cmd is not None:
            setattr(func, "__pre_cmd", pre_cmd)

        if post_cmd is not None:
            setattr(func, "__post_cmd", post_cmd)

        return func

    return func_wrapper


class DynamicCommandMixin:
    """
    This Mixin class defines the functionality to dynamically call specific instrument commands
    from methods that are defined in the Interface classes for device Controllers.

    The mixin uses the `self.transport` instance variables that shall be defined by the
    Controller subclass. The `self.transport` shall be a DeviceTransport object providing the
    methods to read, write, and query an instrument.

    !!! note
        This mixin overrides the `__getattribute__` method!

    !!! note
        This mixin class shall only be inherited from a Controller class that defines the
        `self.transport` attribute.
    """

    def __init__(self):
        if not hasattr(self, "transport"):
            raise AttributeError("Transport was not defined in sub-class of DynamicInterfaceMixin")

    @staticmethod
    def create_command_string(func: Callable, template_str: str, *args, **kwargs):
        """
        Creates a command string that is understood by the instrument. This can be an SCPI
        command or a specific proprietary command string. The `cmd_str` can contain placeholders
        similar to what is used in string formatting.

        As an example, we have a function with two positional arguments 'a', and 'b' and one keyword
        argument flag:

            def func(a, b, flag=True):
                pass

        We have the following template string: `CREATE:FUN:${a} ${b} [${flag}]`.

        When we call the function as follows: `func("TEMP", 23)`, we would then expect
        the returned string to be "CREATE:FUN:TEMP 23 [True]"

            DynamicCommandMixin.create_command_string(func, template, "TEMP", 23)

        Args:
            func (Callable): a function or method that provides the signature
            template_str (str): a template for the command
            args (tuple): positional arguments that will be used in the command string
            kwargs (dict): keywords arguments that will be used in the command string
        """
        try:
            process_kwargs = getattr(func, "__process_kwargs")
        except AttributeError:
            process_kwargs = expand_kwargs

        template = string.Template(template_str)

        sig = inspect.signature(func)
        try:
            bound = sig.bind(*args, **kwargs)
        except TypeError as exc:
            raise CommandError(
                f"Arguments {args}, {kwargs} do not match function signature for "
                f"{func.__name__}{sig}") from exc

        variables = {}
        for idx, par in enumerate(sig.parameters.values()):
            # if the argument is of signature '**kwargs' then expand the kwargs
            if par.kind == inspect.Parameter.VAR_KEYWORD:
                variables[par.name] = process_kwargs(bound.arguments[par.name])
                continue

            # otherwise, use the argument value or the default
            try:
                variables[par.name] = bound.arguments[par.name]
            except KeyError:
                variables[par.name] = par.default

        if hasattr(func, "__use_format"):
            cmd_string = template_str.format(**variables)
        else:
            variables = {
                k: v.value if isinstance(v, enum.Enum) else v
                for k, v in variables.items()
            }
            cmd_string = template.safe_substitute(variables)

        try:
            process_cmd_string = getattr(func, "__process_cmd_string")
            cmd_string = process_cmd_string(cmd_string)
        except AttributeError:
            pass

        return cmd_string

    def handle_dynamic_command(self, attr: Callable) -> Callable:
        """
        Creates a command wrapper calling the appropriate transport methods that are associated
        with the interface definition as passed into this method with the attr argument.

        Args:
            attr: The interface method that is decorated as a dynamic_command.

        Returns:
            Command wrapper with the read or write command, depending on the decorators used
                for that method in the corresponding Interface class.

        Raises:
            AttributeError: If the command is not listed in the YAML file and/or
                has not been listed.
        """

        @functools.wraps(attr)
        def command_wrapper(*args, **kwargs):
            """Generates command strings and executes the transport functions."""
            try:
                cmd_str = getattr(attr, "__cmd_string")
                cmd_str = self.create_command_string(attr, cmd_str, *args, **kwargs)
            except AttributeError:
                cmd_str = None

            response = None

            with contextlib.suppress(AttributeError):
                getattr(attr, "__pre_cmd")(transport=self.transport,
                                           function_name=attr.__name__, cmd_str=cmd_str, args=args, kwargs=kwargs)

            if hasattr(attr, "__write_command"):
                self.transport.write(cmd_str)
            elif hasattr(attr, "__read_command"):
                response = self.transport.read()
            elif hasattr(attr, "__query_command"):
                response = self.transport.query(cmd_str)
            elif hasattr(attr, "__transaction_command"):
                response = self.transport.trans(cmd_str)
            else:
                raise CommandError(f"Interface method '{attr.__name__}' shall be decorated with "
                                   f"a command type decorator.")

            with contextlib.suppress(AttributeError):
                response = getattr(attr, "__post_cmd")(transport=self.transport, response=response)

            with contextlib.suppress(AttributeError):
                process_response = getattr(attr, "__process_response")
                response = process_response(response=response)

            return response

        # Add a hook to identify the command_wrapper function as a method, instead of a function.

        setattr(command_wrapper, "__method_wrapper", True)

        # Add the docstring of the interface method

        command_wrapper.__doc__ = attr.__doc__

        return command_wrapper

    def __getattribute__(self, item):
        """
        The function __getattribute__() is called unconditionally when calling a method or accessing
        an instance variable. This override of `__getattribute__` additionally checks if the
        attribute is a method which has the `__dynamic_interface` defined and then calls a specific
        function to handle the command dynamically.

        Check if item exists:
            - if `item` exists and has the __dynamic_interface attribute then let the function
              handle_dynamic_command() handle this, i.e. call the instrument command.
            - else: the method has been defined in the Controller class, and we should just call
              that method (because it is overridden).
        """

        # If item is not known, an AttributeError will be raised and __getattr__() will be called.

        attr = object.__getattribute__(self, item)

        if hasattr(attr, "__dynamic_interface"):

            # We come here when the method is defined in the Interface class (where it is
            # decorated with the @dynamic_interface), but not in the subclass. So, the method
            # is not overridden. We let the handle_dynamic_command() method handle this.

            attr = self.handle_dynamic_command(attr)

        return attr


class DynamicClientCommandMixin:
    """
    This mixin class contains functionality to forward a device command from a client Proxy class
    to its control server.

    !!! note
        This mixin overrides the `__getattribute__` method!

    """
    def __getattribute__(self, item):
        # If item is not known, an AttributeError will be raised and __getattr__() will be called.

        attr = object.__getattribute__(self, item)

        if hasattr(attr, "__dynamic_interface"):

            # We come here when the method is defined in the Interface class (where it is
            # decorated with the @dynamic_interface), but not in the subclass. So, the method
            # is not overridden. We let the handle_dynamic_command() method handle this.

            attr = self.handle_dynamic_command(attr)

        return attr

    def handle_dynamic_command(self, attr):

        @functools.wraps(attr)
        def command_wrapper(*args, **kwargs):

            attr_name = getattr(attr, "__name__")

            # This will ensure that the function is called with the proper arguments

            try:
                inspect.signature(attr).bind(*args, **kwargs)
            except TypeError as exc:
                LOGGER.error(f"Arguments do not match the signature of the function '{attr_name}': {exc}")
                return None

            # Create a command execution to pass the commanded function and the given arguments
            # to the control server for execution.

            device_method = get_function(self.__class__, attr_name)
            cmd = ClientServerCommand(name=attr_name,
                                      cmd=getattr(attr, "__cmd_string", ""),
                                      response=DynamicCommandProtocol.handle_device_method,
                                      device_method=device_method)
            ce = CommandExecution(cmd, *args, **kwargs)

            # Send the command to the control server for execution

            rc = self.send(ce)
            return rc

        # rewrite the proper signature for the called function

        command_wrapper.__signature__ = inspect.signature(attr)

        return command_wrapper



"""
# Setup

This module defines the Setup, which contains the complete configuration information for a test.

The Setup class contains all configuration items that are specific for a test or observation
and is normally (during nominal operation/testing) loaded automatically from the configuration
manager. The Setup includes type and identification of hardware that is used, calibration files,
software versions, reference frames and coordinate systems that link positions of alignment
equipment, conversion functions for temperature sensors, etc.

The configuration information that is in the Setup can be navigated in two different ways. First,
the Setup is a dictionary, so all information can be accessed by keys as in the following example.

    >>> setup = Setup({"gse": {"hexapod": {"ID": 42, "calibration": [0,1,2,3,4,5]}}})
    >>> setup["gse"]["hexapod"]["ID"]
    42

Second, each of the _keys_ is also available as an attribute of the Setup and that make it
possible to navigate the Setup with dot-notation:

    >>> id = setup.gse.hexapod.ID

In the above example you can see how to navigate from the setup to a device like the PUNA Hexapod.
The Hexapod device is connected to the control server and accepts commands as usual. If you want to
know which keys you can use to navigate the Setup, use the `keys()` method.

    >>> setup.gse.hexapod.keys()
    dict_keys(['ID', 'calibration'])
    >>> setup.gse.hexapod.calibration
    [0, 1, 2, 3, 4, 5]

To get a full printout of the Setup, you can use the `pretty_str()` method. Be careful, because
this can print out a lot of information when a full Setup is loaded.

    >>> print(setup)
    Setup
    └── gse
        └── hexapod
            ├── ID: 42
            └── calibration: [0, 1, 2, 3, 4, 5]

### Special Values

Some of the information in the Setup is interpreted in a special way, i.e. some values are
processed before returning. Examples are the device classes and calibration/data files. The
following values are treated special if they start with:

* `class//`: instantiate the class and return the object
* `factory//`: instantiates a factory and executes its `create()` method
* `csv//`: load the CSV file and return a numpy array
* `yaml//`: load the YAML file and return a dictionary
* `pandas//`: load a CSV file into a pandas Dataframe
* `int-enum//`: dynamically create the enumeration and return the Enum object

#### Device Classes

Most of the hardware components in the Setup will have a `device` key that defines the class for
the device controller. The `device` keys have a value that starts with `class//` and it will
return the device object. As an example, the following defines the Hexapod device:

    >>> setup = Setup(
    ...   {
    ...     "gse": {
    ...       "hexapod": {"ID": 42, "device": "class//egse.hexapod.symetrie.puna.PunaSimulator"}
    ...     }
    ...   }
    ... )
    >>> setup.gse.hexapod.device.is_homing_done()
    False
    >>> setup.gse.hexapod.device.info()
    'Info about the PunaSimulator...'

In the above example you see that we can call the `is_homing_done()` and `info()` methodes
directly on the device by navigating the Setup. It would however be better (more performant) to
put the device object in a variable and work with that variable:

    >>> hexapod = setup.gse.hexapod.device
    >>> hexapod.homing()
    >>> hexapod.is_homing_done()
    True
    >>> hexapod.get_user_positions()

If you need, for some reason, to have access to the actual raw value of the hexapod device key,
use the `get_raw_value()` method:

    >>> setup.gse.hexapod.get_raw_value("device")
    <egse.hexapod.symetrie.puna.PunaSimulator object at ...

#### Data Files

Some information is too large to add to the Setup as such and should be loaded from a data file.
Examples are calibration files, flat-fields, temperature conversion curves, etc.

The Setup will automatically load the file when you access a key that contains a value that
starts with `csv//` or `yaml//`.

    >>> setup = Setup({
    ...     "instrument": {"coeff": "csv//cal_coeff_1234.csv"}
    ... })
    >>> setup.instrument.coeff[0, 4]
    5.0

Note: the resource location is always relative to the path defined by the *PROJECT*_CONF_DATA_LOCATION
environment variable.

The Setup inherits from a NavigableDict (aka navdict) which is also defined in this module.

---

"""

from __future__ import annotations

__all__ = [
    "Setup",
    "navdict",  # noqa: ignore typo
    "list_setups",
    "load_setup",
    "get_setup",
    "submit_setup",
    "SetupError",
    "load_last_setup_id",
    "save_last_setup_id",
]

import enum
import importlib
import logging
import os
import re
import textwrap
import warnings
from functools import lru_cache
from pathlib import Path
from typing import Any
from typing import Optional
from typing import Union

import rich
import yaml
from rich.tree import Tree

from egse.env import get_conf_data_location
from egse.env import get_conf_repo_location
from egse.env import get_conf_repo_location_env_name
from egse.env import get_data_storage_location
from egse.env import has_conf_repo_location
from egse.env import print_env
from egse.response import Failure
from egse.settings import read_configuration_file
from egse.system import format_datetime
from egse.system import sanity_check
from egse.system import walk_dict_tree

MODULE_LOGGER = logging.getLogger(__name__)


class SetupError(Exception):
    """ A setup-specific error."""
    pass


def _load_class(class_name: str):
    """
    Find and returns a class based on the fully qualified name.

    A class name can be preceded with the string `class//` or `factory//`. This is used in YAML
    files where the class is then instantiated on load.

    Args:
        class_name (str): a fully qualified name for the class
    """
    if class_name.startswith("class//"):
        class_name = class_name[7:]
    elif class_name.startswith("factory//"):
        class_name = class_name[9:]

    module_name, class_name = class_name.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def _load_csv(resource_name: str):
    """Find and return the content of a CSV file."""
    from numpy import genfromtxt  # FIXME: use CSV standard module

    parts = resource_name[5:].rsplit("/", 1)
    [in_dir, fn] = parts if len(parts) > 1 else [None, parts[0]]
    conf_location = get_conf_data_location()
    try:
        csv_location = Path(conf_location) / in_dir / fn
        content = genfromtxt(csv_location, delimiter=",", skip_header=1)
    except TypeError as exc:
        raise ValueError(
            f"Couldn't load resource '{resource_name}' from default {conf_location=}") from exc
    return content


def _load_int_enum(enum_name: str, enum_content):
    """ Dynamically build (and return) and IntEnum.

    Args:
        - enum_name: Enumeration name (potentially prepended with "int_enum//").
        - enum_content: Content of the enumeration, as read from the setup.
    """
    if enum_name.startswith("int_enum//"):
        enum_name = enum_name[10:]

    definition = {}
    for side_name, side_definition in enum_content.items():

        if "alias" in side_definition:
            aliases = side_definition["alias"]
        else:
            aliases = []
        value = side_definition["value"]

        definition[side_name] = value

        for alias in aliases:
            definition[alias] = value
    return enum.IntEnum(enum_name, definition)


def _load_yaml(resource_name: str):
    """Find and return the content of a YAML file."""
    from egse.settings import Settings
    from egse.settings import SettingsError

    parts = resource_name[6:].rsplit("/", 1)
    [in_dir, fn] = parts if len(parts) > 1 else [None, parts[0]]
    conf_location = get_conf_data_location()
    try:
        yaml_location = Path(conf_location) / in_dir
        content = NavigableDict(Settings.load(location=yaml_location, filename=fn, add_local_settings=False))
    except (TypeError, SettingsError) as exc:
        raise ValueError(
            f"Couldn't load resource '{resource_name}' from default {conf_location=}") from exc
    return content


def _load_pandas(resource_name: str, separator: str):
    """
    Find and return the content of the given file as a pandas DataFrame object.

    The file is loaded relative from the location of the configuration data
    as defined by `get_conf_data_location()`.

    Args:
        - resource_name: Filename, preceded by "pandas//".
        - separator: Column separator.
    """
    import pandas

    parts = resource_name[8:].rsplit("/", 1)
    [in_dir, fn] = parts if len(parts) > 1 else [None, parts[0]]
    conf_location = get_conf_data_location()

    try:
        pandas_file_location = Path(conf_location) / in_dir / fn
        return pandas.read_csv(pandas_file_location, sep=separator)
    except TypeError as exc:
        raise ValueError(
            f"Couldn't load resource '{resource_name}' from default {conf_location=}") from exc


def _get_attribute(self, name, default):
    try:
        attr = object.__getattribute__(self, name)
    except AttributeError:
        attr = default
    return attr


def _parse_filename_for_setup_id(filename: str) -> str | None:
    """Returns the setup_id from the filename, or None when no match was found."""

    # match = re.search(r"SETUP_([^_]+)_(\d+)", filename)
    match = re.search(r"SETUP_(\w+)_([\d]{5})_([\d]{6})_([\d]{6})\.yaml", filename)

    # TypeError when match is None

    try:
        return match[2]  # match[2] is setup_id
    except (IndexError, TypeError):
        return None


def disentangle_filename(filename: str) -> tuple:
    """
    Returns the site_id and setup_id (as a tuple) that is extracted from the Setups filename.

    Args:
        filename (str): the filename or fully qualified file path as a string.

    Returns:
        A tuple (site_id, setup_id).
    """
    if filename is None:
        return ()

    match = re.search(r"SETUP_(\w+)_([\d]{5})_([\d]{6})_([\d]{6})\.yaml", filename)

    if match is None:
        return ()

    site_id, setup_id = match[1], match[2]

    return site_id, setup_id


def get_last_setup_id_file_path(site_id: str = None) -> Path:
    """
    Return the fully expanded file path of the file containing the last loaded Setup in the configuration manager.
    The default location for this file is the data storage location.

    Args:
        site_id: The SITE identifier (overrides the SITE_ID environment variable)

    """
    location = get_data_storage_location(site_id=site_id)

    return Path(location).expanduser().resolve() / "last_setup_id.txt"


def load_last_setup_id(site_id: str = None) -> int:
    """
    Returns the ID of the last Setup that was used by the configuration manager.
    The file shall only contain the Setup ID which must be an integer on the first line of the file.
    If no such ID can be found, the Setup ID = 0 will be returned.

    Args:
        site_id: The SITE identifier
    """

    last_setup_id_file_path = get_last_setup_id_file_path(site_id=site_id)
    try:
        with last_setup_id_file_path.open('r') as fd:
            setup_id = int(fd.read().strip())
    except FileNotFoundError:
        setup_id = 0
        save_last_setup_id(setup_id)

    return setup_id


def save_last_setup_id(setup_id: int | str, site_id: str = None):
    """
    Makes the given Setup ID persistent, so it can be restored upon the next startup.

    Args:
        setup_id: The Setup identifier to be saved
        site_id: The SITE identifier

    """

    last_setup_id_file_path = get_last_setup_id_file_path(site_id=site_id)
    with last_setup_id_file_path.open('w') as fd:
        fd.write(f"{int(setup_id):d}")


class NavigableDict(dict):
    """
    A NavigableDict is a dictionary where all keys in the original dictionary are also accessible
    as attributes to the class instance. So, if the original dictionary (setup) has a key
    "site_id" which is accessible as `setup['site_id']`, it will also be accessible as
    `setup.site_id`.

    Examples:
        >>> setup = NavigableDict({'site_id': 'KU Leuven', 'version': "0.1.0"})
        >>> assert setup['site_id'] == setup.site_id
        >>> assert setup['version'] == setup.version

    Note:
        We always want **all** keys to be accessible as attributes, or none. That means all
        keys of the original dictionary shall be of type `str`.

    """

    def __init__(self, head: dict = None, label: str = None):
        """
        Args:
            head (dict): the original dictionary
            label (str): a label or name that is used when printing the navdict
        """
        head = head or {}
        super().__init__(head)
        self.__dict__["_memoized"] = {}
        self.__dict__["_label"] = label

        # By agreement, we only want the keys to be set as attributes if all keys are strings.
        # That way we enforce that always all keys are navigable, or none.

        if any(True for k in head.keys() if not isinstance(k, str)):
            return

        for key, value in head.items():
            if isinstance(value, dict):
                setattr(self, key, NavigableDict(head.__getitem__(key)))
            else:
                setattr(self, key, head.__getitem__(key))

    def add(self, key: str, value: Any):
        """Set a value for the given key.

        If the value is a dictionary, it will be converted into a NavigableDict and the keys
        will become available as attributes provided that all the keys are strings.

        Args:
            key (str): the name of the key / attribute to access the value
            value (Any): the value to assign to the key
        """
        if isinstance(value, dict) and not isinstance(value, NavigableDict):
            value = NavigableDict(value)
        setattr(self, key, value)

    def clear(self) -> None:
        for key in list(self.keys()):
            self.__delitem__(key)

    def __repr__(self):
        return f"{self.__class__.__name__}({super()!r})"

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        object.__delattr__(self, key)

    def __setattr__(self, key, value):
        # MODULE_LOGGER.info(f"called __setattr__({self!r}, {key}, {value})")
        if isinstance(value, dict) and not isinstance(value, NavigableDict):
            value = NavigableDict(value)
        self.__dict__[key] = value
        super().__setitem__(key, value)
        try:
            del self.__dict__["_memoized"][key]
        except KeyError:
            pass

    def __getattribute__(self, key):
        # MODULE_LOGGER.info(f"called __getattribute__({key})")
        value = object.__getattribute__(self, key)
        if isinstance(value, str) and value.startswith("class//"):
            try:
                dev_args = object.__getattribute__(self, 'device_args')
            except AttributeError:
                dev_args = ()
            return _load_class(value)(*dev_args)
        if isinstance(value, str) and value.startswith("factory//"):
            factory_args = _get_attribute(self, f'{key}_args', {})
            return _load_class(value)().create(**factory_args)
        if isinstance(value, str) and value.startswith("int_enum//"):
            content = object.__getattribute__(self, "content")
            return _load_int_enum(value, content)
        if isinstance(value, str) and value.startswith("csv//"):
            if key in self.__dict__["_memoized"]:
                return self.__dict__["_memoized"][key]
            content = _load_csv(value)
            self.__dict__["_memoized"][key] = content
            return content
        if isinstance(value, str) and value.startswith("yaml//"):
            if key in self.__dict__["_memoized"]:
                return self.__dict__["_memoized"][key]
            content = _load_yaml(value)
            self.__dict__["_memoized"][key] = content
            return content
        if isinstance(value, str) and value.startswith("pandas//"):
            separator = object.__getattribute__(self, 'separator')
            return _load_pandas(value, separator)
        else:
            return value

    def __delattr__(self, item):
        # MODULE_LOGGER.info(f"called __delattr__({self!r}, {item})")
        object.__delattr__(self, item)
        dict.__delitem__(self, item)

    def __setitem__(self, key, value):
        # MODULE_LOGGER.info(f"called __setitem__({self!r}, {key}, {value})")
        if isinstance(value, dict) and not isinstance(value, NavigableDict):
            value = NavigableDict(value)
        super().__setitem__(key, value)
        self.__dict__[key] = value
        try:
            del self.__dict__["_memoized"][key]
        except KeyError:
            pass

    def __getitem__(self, key):
        # MODULE_LOGGER.info(f"called __getitem__({self!r}, {key})")
        value = super().__getitem__(key)
        if isinstance(value, str) and value.startswith("class//"):
            try:
                dev_args = object.__getattribute__(self, 'device_args')
            except AttributeError:
                dev_args = ()
            return _load_class(value)(*dev_args)
        if isinstance(value, str) and value.startswith("csv//"):
            return _load_csv(value)
        if isinstance(value, str) and value.startswith("int_enum//"):
            content = object.__getattribute__(self, "content")
            return _load_int_enum(value, content)
        else:
            return value

    def set_private_attribute(self, key: str, value: Any) -> None:
        """Sets a private attribute for this object.

        The name in key will be accessible as an attribute for this object, but the key will not
        be added to the dictionary and not be returned by methods like keys().

        The idea behind this private attribute is to have the possibility to add status information
        or identifiers to this classes object that can be used by save() or load() methods.

        Args:
            key (str): the name of the private attribute (must start with an underscore character).
            value: the value for this private attribute

        Examples:
            >>> setup = NavigableDict({'a': 1, 'b': 2, 'c': 3})
            >>> setup.set_private_attribute("_loaded_from_dict", True)
            >>> assert "c" in setup
            >>> assert "_loaded_from_dict" not in setup
            >>> assert setup.get_private_attribute("_loaded_from_dict") == True

        """
        if key in self:
            raise ValueError(
                f"Invalid argument key='{key}', this key already exists in dictionary."
            )
        if not key.startswith("_"):
            raise ValueError(
                f"Invalid argument key='{key}', must start with underscore character '_'."
            )
        self.__dict__[key] = value

    def get_private_attribute(self, key: str) -> Any:
        """Returns the value of the given private attribute.

        Args:
            key (str): the name of the private attribute (must start with an underscore character).

        Returns:
            the value of the private attribute given in `key`.

        Note:
            Because of the implementation, this private attribute can also be accessed as a 'normal'
            attribute of the object. This use is however discouraged as it will make your code less
            understandable. Use the methods to access these 'private' attributes.
        """
        if not key.startswith("_"):
            raise ValueError(
                f"Invalid argument key='{key}', must start with underscore character '_'."
            )
        return self.__dict__[key]

    def has_private_attribute(self, key):
        """
        Check if the given key is defined as a private attribute.

        Args:
            key (str): the name of a private attribute (must start with an underscore)
        Returns:
            True if the given key is a known private attribute.
        Raises:
            ValueError: when the key doesn't start with an underscore.
        """
        if not key.startswith("_"):
            raise ValueError(f"Invalid argument key='{key}', must start with underscore character '_'.")

        try:
            _ = self.__dict__[key]
            return True
        except KeyError:
            return False

    def get_raw_value(self, key):
        """
        Returns the raw value of the given key.

        Some keys have special values that are interpreted by the AtributeDict class. An example is
        a value that starts with 'class//'. When you access these values, they are first converted
        from their raw value into their expected value, e.g. the instantiated object in the above
        example. This method allows you to access the raw value before conversion.
        """
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            raise KeyError(f"The key '{key}' is not defined.")

    def __str__(self):
        return self.pretty_str()

    def pretty_str(self, indent: int = 0):
        """
        Returns a pretty string representation of the dictionary.

        Args:
            indent (int): number of indentations (of four spaces)

        Note:
            The indent argument is intended for the recursive call of this function.
        """
        msg = ""

        for k, v in self.items():
            if isinstance(v, NavigableDict):
                msg += f"{'    '*indent}{k}:\n"
                msg += v.pretty_str(indent + 1)
            else:
                msg += f"{'    '*indent}{k}: {v}\n"

        return msg

    def __rich__(self) -> Tree:
        tree = Tree(self.__dict__["_label"] or "NavigableDict", guide_style="dim")
        walk_dict_tree(self, tree, text_style="dark grey")
        return tree

    def _save(self, fd, indent: int = 0):
        """
        Recursive method to write the dictionary to the file descriptor.

        Indentation is done in steps of four spaces, i.e. `'    '*indent`.

        Args:
            fd: a file descriptor as returned by the open() function
            indent (int): indentation level of each line [default = 0]

        """
        from egse.device import DeviceInterface

        # Note that the .items() method returns the actual values of the keys and doesn't use the
        # __getattribute__ or __getitem__ methods. So the raw value is returned and not the
        # _processed_ value.

        for k, v in self.items():

            # history shall be saved last, skip it for now

            if k == "history":
                continue

            # make sure to escape a colon in the key name

            if isinstance(k, str) and ":" in k:
                k = '"' + k + '"'

            if isinstance(v, NavigableDict):
                fd.write(f"{'    '*indent}{k}:\n")
                v._save(fd, indent + 1)
                fd.flush()
                continue

            if isinstance(v, DeviceInterface):
                v = f"class//{v.__module__}.{v.__class__.__name__}"
            if isinstance(v, float):
                v = f"{v:.6E}"
            fd.write(f"{'    '*indent}{k}: {v}\n")
            fd.flush()

        # now save the history as the last item

        if "history" in self:
            fd.write(f"{'    ' * indent}history:\n")
            self.history._save(fd, indent + 1)

    def get_memoized_keys(self):
        return list(self.__dict__["_memoized"].keys())


navdict = NavigableDict  # noqa: ignore typo
"""Shortcut for NavigableDict and more Pythonic."""


class Setup(NavigableDict):
    """The Setup class represents a version of the configuration of the test facility, the
    test setup and the Camera Under Test (CUT)."""

    def __init__(self, nav_dict: NavigableDict | dict = None, label: str = None):
        super().__init__(nav_dict or {}, label=label)

    @staticmethod
    def from_dict(my_dict):
        """Create a Setup from a given dictionary.

        Remember that all keys in the given dictionary shall be of type 'str' in order to be
        accessible as attributes.

        Examples:
            >>> setup = Setup.from_dict({"ID": "my-setup-001", "version": "0.1.0"})
            >>> assert setup["ID"] == setup.ID == "my-setup-001"

        """
        return Setup(my_dict, label="Setup")

    @staticmethod
    def from_yaml_string(yaml_content: str = None) -> Setup:
        """Loads a Setup from the given YAML string.

        This method is mainly used for easy creation of Setups from strings during unit tests.

        Args:
            yaml_content (str): a string containing YAML

        Returns:
            a Setup that was loaded from the content of the given string.
        """

        if not yaml_content:
            raise ValueError("Invalid argument to function: No input string or None given.")

        setup_dict = yaml.safe_load(yaml_content)

        if "Setup" in setup_dict:
            setup_dict = setup_dict["Setup"]

        return Setup(setup_dict, label="Setup")

    @staticmethod
    @lru_cache(maxsize=300)
    def from_yaml_file(filename: Union[str, Path] = None, add_local_settings: bool = True) -> Setup:
        """Loads a Setup from the given YAML file.

        Args:
            filename (str): the path of the YAML file to be loaded
            add_local_settings (bool): if local settings shall be loaded and override the settings from the YAML file.

        Returns:
            a Setup that was loaded from the given location.

        Raises:
            ValueError: when no filename is given.
        """

        if not filename:
            raise ValueError("Invalid argument to function: No filename or None given.")

        # MODULE_LOGGER.info(f"Loading {filename}...")

        setup_dict = read_configuration_file(filename, force=True)
        if setup_dict == {}:
            warnings.warn(f"Empty Setup file: {filename!s}")

        try:
            setup_dict = setup_dict["Setup"]
        except KeyError:
            warnings.warn(f"Setup file doesn't have a 'Setup' group: {filename!s}")

        setup = Setup(setup_dict, label="Setup")
        setup.set_private_attribute("_filename", Path(filename))
        if setup_id := _parse_filename_for_setup_id(str(filename)):
            setup.set_private_attribute("_setup_id", setup_id)

        return setup

    def to_yaml_file(self, filename=None):
        """Saves a NavigableDict to a YAML file.

        When no filename is provided, this method will look for a 'private' attribute
        `_filename` and use that to save the data.

        Args:
            filename (str|Path): the path of the YAML file where to save the data

        Note:
            This method will **overwrite** the original or given YAML file and therefore you might
            lose proper formatting and/or comments.

        """
        if not filename:
            try:
                filename = self.get_private_attribute("_filename")
            except KeyError:
                raise ValueError("No filename given or known, can not save Setup.")

        print(f"Saving Setup to {filename!s}")

        with Path(filename).open("w") as fd:

            fd.write(
                f"# Setup generated by:\n"
                f"#\n"
                f"#    Setup.to_yaml_file(setup, filename='{filename}')\n#\n"
            )
            fd.write(f"# Created on {format_datetime()}\n\n")
            fd.write("Setup:\n")

            self._save(fd, indent=1)

        self.set_private_attribute("_filename", Path(filename))

    @staticmethod
    def compare(setup_1: NavigableDict, setup_2: NavigableDict):
        from egse.device import DeviceInterface
        from deepdiff import DeepDiff

        return DeepDiff(setup_1, setup_2, exclude_types=[DeviceInterface])

    # def get_devices(self):
    #     """Returns a list of devices for the current setup.

    #     Returns:
    #         - List of devices for the current setup.
    #     """

    #     devices = []

    #     Setup.walk(self, "device", devices)

    #     return devices

    @staticmethod
    def find_devices(node: NavigableDict, devices: dict = None) -> dict[str, tuple[str, tuple]]:
        """
        Returns a dictionary with the devices that are included in the setup.  The keys
        in the dictionary are taken from the "device_name" entries in the setup file. The
        corresponding values in the dictionary are taken from the "device" entries in the
        setup file.

        Args:
            node: Dictionary in which to look for the devices (and their names).
            devices: Dictionary in which to include the devices in the setup.

        Returns:
            Dictionary with the devices that are included in the setup. The keys are the device name,
                the values are tuples with the 'device' raw value and the device arguments as a tuple.
        """
        devices = devices or {}

        for sub_node in node.values():

            if isinstance(sub_node, NavigableDict):

                if ("device" in sub_node) and ("device_name" in sub_node):

                    device = sub_node.get_raw_value("device")

                    if "device_args" in sub_node:
                        device_args = sub_node.get_raw_value("device_args")
                    else:
                        device_args = ()

                    devices[sub_node["device_name"]] = (device, device_args)

                else:

                    Setup.find_devices(sub_node, devices=devices)

        return devices

    @staticmethod
    def walk(node: dict, key_of_interest: str, leaf_list: list) -> list:

        """
        Walk through the given dictionary, in a recursive way, appending the leaf with
        the given keyword to the given list.

        Args:
            node: Dictionary in which to look for leaves with the given keyword.
            key_of_interest: Key to look for in the leaves of the given dictionary.
            leaf_list: List to which to add the leaves with the given keyword.

        Returns:
            Given list with the leaves (with the given keyword) in the given dictionary \
            appended to it.
        """

        for key, sub_node in node.items():

            if isinstance(sub_node, dict):

                Setup.walk(sub_node, key_of_interest, leaf_list)

            elif key == key_of_interest:

                leaf_list.append(sub_node)

        return leaf_list

    def __rich__(self) -> Tree:
        tree = super().__rich__()
        if self.has_private_attribute("_setup_id"):
            setup_id = self.get_private_attribute('_setup_id')
            tree.add(f"Setup ID: {setup_id}", style="grey50")
        if self.has_private_attribute("_filename"):
            filename = self.get_private_attribute('_filename')
            tree.add(f"Loaded from: {filename}", style="grey50")
        return tree

    def get_id(self) -> Optional[str]:
        """Returns the Setup ID (as a string) or None when no setup id could be identified."""
        if self.has_private_attribute("_setup_id"):
            return self.get_private_attribute('_setup_id')
        else:
            return None

    def get_filename(self) -> Optional[str]:
        """Returns the filename for this Setup or None when no filename could be determined."""
        if self.has_private_attribute("_filename"):
            return self.get_private_attribute('_filename')
        else:
            return None


def list_setups(**attr):
    """
    This is a function to be used for interactive use, it will print to the terminal (stdout) a
    list of Setups known at the Configuration Manager. This list is sorted with the most recent (
    highest) value last.

    The list can be restricted with key:value pairs (keyword arguments). This _search_ mechanism
    allows us to find all Setups that adhere to the key:value pairs, e.g. to find all Setups for
    CSL at position 2, use:

        >>> list_setups(site_id="CSL", position=2)

    To have a nested keyword search (i.e. search by `gse.hexapod.ID`) then pass in
    `gse__hexapod__ID` as the keyword argument. Replace the '.' notation with double underscores
    '__'.

        >>> list_setups(gse__hexapod__ID=4)
    """

    try:
        from egse.confman import ConfigurationManagerProxy
    except ImportError:
        print("WARNING: package 'cgse-core' is not installed, service not available.")
        return

    try:
        with ConfigurationManagerProxy() as proxy:
            setups = proxy.list_setups(**attr)
        if setups:
            # We want to have the most recent (highest id number) last, but keep the site together
            setups = sorted(setups, key=lambda x: (x[1], x[0]))
            print("\n".join(f"{setup}" for setup in setups))
        else:
            print("no Setups found")
    except ConnectionError:
        print("Could not make a connection with the Configuration Manager, no Setup to show you.")


def get_setup(setup_id: int = None):
    """
    Retrieve the currently active Setup from the configuration manager.

    When a setup_id is provided, that setup will be returned, but not loaded in the configuration
    manager. This function does NOT change the configuration manager.

    This function is for interactive use and consults the configuration manager server. Don't use
    this within the test script, but use the `GlobalState.setup` property instead.
    """
    try:
        from egse.confman import ConfigurationManagerProxy
    except ImportError:
        print("WARNING: package 'cgse-core' is not installed, service not available.")
        return

    try:
        with ConfigurationManagerProxy() as proxy:
            setup = proxy.get_setup(setup_id)
        return setup
    except ConnectionError:
        print("Could not make a connection with the Configuration Manager, no Setup returned.")


def _check_conditions_for_get_path_of_setup_file(site_id: str) -> Path:
    """
    Check some pre-conditions that need to be met before we try to determine the
    file path for the requested Setup file.

    The following checks are performed:

    * if the environment variable '{PROJECT}_CONF_REPO_LOCATION' is set
    * if the directory specified in the env variable actually exists
    * if the folder with the Setups exists for the given site_id

    Args:
        site_id (str): the name of the test house

    Returns:
        The location of the Setup files for the given test house.

    Raises:
        LookupError when the environment variable is not set.

        NotADirectoryError when either the repository folder or the Setups folder doesn't exist.

    """
    repo_location_env = get_conf_repo_location_env_name()

    if not (repo_location := get_conf_repo_location()):
        raise LookupError(
            f"Environment variable doesn't exist or points to an invalid location, please (re-)define"
            f" {repo_location_env} and try again."
        )

    print_env()

    repo_location = Path(repo_location)
    setup_location = repo_location / 'data' / site_id / 'conf'

    if not repo_location.is_dir():
        raise NotADirectoryError(
            f"The location of the repository for Setup files doesn't exist: {repo_location!s}. "
            f"Please check the environment variable {repo_location_env}."
        )

    if not setup_location.is_dir():
        raise NotADirectoryError(
            f"The location of the Setup files doesn't exist: {setup_location!s}. "
            f"Please check if the given {site_id=} is correct."
        )

    return setup_location


def get_path_of_setup_file(setup_id: int, site_id: str) -> Path:
    """
    Returns the Path to the last Setup file for the given site_id. The last Setup
    file is the file with the largest setup_id number.

    This function needs the environment variable <PROJECT>_CONF_REPO_LOCATION to
    be defined as the location of the repository with configuration data on your
    disk. If the repo is not defined, the configuration data location will be used
    instead.

    Args:
        setup_id (int): the identifier for the requested Setup
        site_id (str): the test house name, one of CSL, SRON, IAS, INTA

    Returns:
        The full path to the requested Setup file.

    Raises:
        LookupError: when the environment variable is not set.

        NotADirectoryError: when either the repository folder or the Setups folder doesn't exist.

        FileNotFoundError: when no Setup file can be found for the given arguments.

    """

    if not has_conf_repo_location():
        setup_location = Path(get_conf_data_location(site_id))
    else:
        setup_location = _check_conditions_for_get_path_of_setup_file(site_id)

    if setup_id:
        files = list(setup_location.glob(f'SETUP_{site_id}_{setup_id:05d}_*.yaml'))

        if not files:
            raise FileNotFoundError(f"No Setup found for {setup_id=} and {site_id=}.")

        file_path = Path(setup_location) / files[-1]
    else:
        files = setup_location.glob('SETUP*.yaml')

        last_file_parts = sorted([file.name.split('_') for file in files])[-1]
        file_path = Path(setup_location) / "_".join(last_file_parts)

    sanity_check(file_path.is_file(), f"The expected Setup file doesn't exist: {file_path!s}")

    return file_path


def load_setup(
        setup_id: int = None,
        site_id: str = None, from_disk: bool = False) -> Setup:
    """
    This function loads the Setup corresponding with the given `setup_id`.

    Loading a Setup means:

    * that this Setup will also be loaded and activated in the configuration manager,
    * that this Setup will be available from the `GlobalState.setup`

    When no setup_id is provided, the current Setup is loaded from the configuration manager.

    Args:
        setup_id (int): the identifier for the Setup
        site_id (str): the name of the test house
        from_disk (bool): True if the Setup needs to be loaded from disk

    Returns:
        The requested Setup or None when the Setup could not be loaded from the \
        configuration manager.

    """
    from egse.state import GlobalState

    if from_disk:
        if site_id is None:
            raise ValueError(
                "The site_id argument can not be empty when from_disk is given and True")

        setup_file_path = get_path_of_setup_file(setup_id, site_id)

        rich.print(
            f"Loading {'' if setup_id else 'the latest '}Setup {f'{setup_id} ' if setup_id else ''}for {site_id}..."
        )

        return Setup.from_yaml_file(setup_file_path)

    # When we arrive here the Setup shall be loaded from the Configuration manager

    try:
        from egse.confman import ConfigurationManagerProxy
    except ImportError:
        print("WARNING: package 'cgse-core' is not installed, service not available. Returning an empty Setup.")
        return Setup()

    if setup_id is not None:
        try:
            with ConfigurationManagerProxy() as proxy:
                proxy.load_setup(setup_id)

        except ConnectionError:
            MODULE_LOGGER.warning(
                "Could not make a connection with the Configuration Manager, no Setup to show you."
            )
            rich.print(
                "\n"
                "If you are not running this from an operational machine, do not have a CM "
                "running locally or don't know what this means, then: \n"
                "  (1) define the environment variable 'PLATO_CONF_REPO_LOCATION' and \n"
                "      it points to the location of the plato-cgse-conf repository,\n"
                "  (2) try again using the argument 'from_disk=True'.\n"
            )

    return GlobalState.load_setup()


def submit_setup(setup: Setup, description: str) -> str | None:
    """
    Submit the given Setup to the Configuration Manager.

    When you submit a Setup, the Configuration Manager will save this Setup with the
    next (new) setup id and make this Setup the current Setup in the Configuration manager
    unless you have explicitly set `replace=False` in which case the current Setup will
    not be replaced with the new Setup.

    Args:
        setup (Setup): a (new) Setup to submit to the configuration manager
        description (str): one-liner to help identifying the Setup afterwards

    Returns:
        The Setup ID of the newly created Setup or None.
    """
    # We have not yet decided if this option should be made available. Therefore, we
    # leave it here as hardcoded True.

    # replace (bool): True if the current Setup in the configuration manager shall
    #                 be replaced by this new Setup. [default=True]
    replace: bool = True

    try:
        from egse.confman import ConfigurationManagerProxy
    except ImportError:
        print("WARNING: package 'cgse-core' is not installed, service not available.")
        return

    try:
        with ConfigurationManagerProxy() as proxy:
            setup = proxy.submit_setup(setup, description, replace)

        if setup is None:
            rich.print("[red]Submit failed for given Setup, no reason given.[/red]")
        elif isinstance(setup, Failure):
            rich.print(f"[red]Submit failed for given Setup[/red]: {setup}")
            setup = None
        elif replace:
            rich.print(
                textwrap.dedent(
                    """\
                    [green]
                    Your new setup has been submitted and pushed to GitHub. The new setup is also
                    activated in the configuration manager. Load the new setup in your session with:

                        setup = load_setup()
                    [/]
                    """
                )
            )
        else:
            rich.print(
                textwrap.dedent(
                    """[dark_orange]
                    Your new setup has been submitted and pushed to GitHub, but has not been
                    activated in the configuration manager. To activate this setup, use the
                    following command:

                        setup = load_setup({str(setup.get_id())})
                    [/]
                    """
                )
            )

        return setup.get_id() if setup is not None else None

    except ConnectionError:
        rich.print("Could not make a connection with the Configuration Manager, no Setup was submitted.")
    except NotImplementedError:
        rich.print(
            textwrap.dedent(
                """\
                Caught a NotImplementedError. That usually means the configuration manager is not running or
                can not be reached. Check on the egse-server if the `cm_cs` process is running. If not you will
                need to be restart the core services.
                """
            )
        )


def main(args: list = None):  # pragma: no cover
    import argparse

    from rich import print

    from egse.config import find_files
    from egse.settings import Settings

    SITE = Settings.load("SITE")
    location = os.environ.get("PLATO_CONF_DATA_LOCATION")
    parser = argparse.ArgumentParser(
        description=textwrap.dedent("""\
            Print out the Setup for the given setup-id. The Setup will
            be loaded from the location given by the environment variable
            PLATO_CONF_DATA_LOCATION. If this env is not set, the Setup
            will be searched from the current directory."""
        ),
        epilog=f"PLATO_CONF_DATA_LOCATION={location}"
    )
    parser.add_argument(
        "--setup-id", type=int, default=-1,
        help="the Setup ID. If not given, the last Setup will be selected.")
    parser.add_argument("--list", "-l", action="store_true", help="list available Setups.")
    parser.add_argument("--use-cm", action="store_true", help="use the configuration manager.")
    args = parser.parse_args(args or [])

    if args.use_cm:
        try:
            from egse.confman import ConfigurationManagerProxy
        except ImportError:
            print("WARNING: package 'cgse-core' is not installed, service not available.")
            return

        with ConfigurationManagerProxy() as cm:
            if args.list:
                print(cm.list_setups())
            else:
                print(cm.get_setup())
        return

    if args.list:
        files = find_files(f"SETUP_{SITE.ID}_*_*.yaml", root=location)
        files = list(files)
        if files:
            location = files[0].parent.resolve()
        print(sorted([f.name for f in files]))
        print(f"Loaded from [purple]{location}.")
    else:
        setup_id = args.setup_id
        if setup_id == -1:
            setup_files = find_files(f"SETUP_{SITE.ID}_*_*.yaml", root=location)
        else:
            setup_files = find_files(f"SETUP_{SITE.ID}_{setup_id:05d}_*.yaml", root=location)
        setup_files = list(setup_files)
        if len(setup_files) > 0:
            setup_file = sorted(setup_files)[-1]
            setup = Setup.from_yaml_file(setup_file)
            print(setup)
        else:
            print("[red]No setup files were found.[/]")


if __name__ == "__main__":
    import sys
    main(sys.argv[1:])



import pickle
import zlib
from enum import IntEnum


def connect_address(transport, address, port):
    """Returns a properly formatted URL to connect to."""
    return f"{transport}://{address}:{port}"


def bind_address(transport, port):
    """Returns a properly formatted url to bind a socket to."""
    return f"{transport}://*:{port}"


def set_address_port(url: str, port: int):
    """Returns a url where the 'port' part is replaced with the given port."""
    transport, address, old_port = split_address(url)

    return f"{transport}://{address}:{port}"


def split_address(url: str):

    transport, address, port = url.split(":")
    if address.startswith("//"):
        address = address[2:]
    return transport, address, port


def send_zipped_pickle(socket, obj, flags=0, protocol=-1):
    """pickle an object, and zip the pickle before sending it"""
    p = pickle.dumps(obj, protocol)
    z = zlib.compress(p)
    return socket.send(z, flags=flags)


def recv_zipped_pickle(socket, flags=0, protocol=-1):
    """inverse of send_zipped_pickle"""
    z = socket.recv(flags)
    p = zlib.decompress(z)
    return pickle.loads(p)


class MessageIdentifier(IntEnum):
    """
    The first item in a multipart message that can be used to subscribe, filter and identify
    messages.
    """

    # ALL shall not be used in the multipart message itself, but exists as an indicator for
    # subscribing to all messages. The ALL shall be converted into b'' when subscribing.

    ALL = 0x00

    # Synchronisation to DPU Processor at time of reception

    SYNC_TIMECODE = 0x80
    SYNC_HK_PACKET = 0x81
    SYNC_DATA_PACKET = 0x82
    SYNC_ERROR_FLAGS = 0x85
    SYNC_HK_DATA = 0x86

    N_FEE_REGISTER_MAP = 0x83
    NUM_CYCLES = 0x84

    # Sending out all kinds of information

    HDF5_FILENAMES = 0x90



import contextlib
import re
import subprocess
import sys
import textwrap
from pathlib import Path
from typing import Annotated

import rich
import typer

from egse.plugin import entry_points
from egse.system import format_datetime

app = typer.Typer()


@app.command()
def top():
    """
    A top-like interface for core services and device control servers.

    Not yet implemented.
    """
    print("This fancy top is not yet implemented.")


@app.command()
def init(project: str = ""):
    """Initialize your project."""
    from rich.prompt import Prompt, Confirm

    project = project.upper()
    site_id = None

    rich.print("[light_steel_blue]Please note default values are given between \[brackets].[/]")

    answer = Prompt.ask(f"What is the name of the project [{project}] ?")
    if answer:
        project = answer.upper()
    while not site_id:
        answer = Prompt.ask("What is the site identifier ?")
        if answer:
            site_id = answer.upper()
        else:
            answer = Confirm.ask("Abort?")
            if answer:
                return

    data_storage_location = f"~/data/{project}/{site_id}/"
    answer = Prompt.ask(f"Where can the project data be stored [{data_storage_location}] ?")
    if answer:
        data_storage_location = answer

    conf_data_location = f"~/data/{project}/{site_id}/conf/"
    answer = Prompt.ask(f"Where will the configuration data be located [{conf_data_location}] ?")
    if answer:
        conf_data_location = answer

    log_file_location = f"~/data/{project}/{site_id}/log/"
    answer = Prompt.ask(f"Where will the logging messages be stored [{log_file_location}] ?")
    if answer:
        log_file_location = answer

    local_settings_path = f"~/data/{project}/{site_id}/local_settings.yaml"
    answer = Prompt.ask(f"Where shall I create a local settings YAML file [{local_settings_path}] ?")
    if answer:
        local_settings_path = answer

    Path(data_storage_location).expanduser().mkdir(exist_ok=True, parents=True)
    (Path(data_storage_location).expanduser() / "daily").mkdir(exist_ok=True, parents=True)
    (Path(data_storage_location).expanduser() / "obs").mkdir(exist_ok=True, parents=True)
    Path(conf_data_location).expanduser().mkdir(exist_ok=True, parents=True)
    Path(log_file_location).expanduser().mkdir(exist_ok=True, parents=True)

    with open(Path(local_settings_path).expanduser(), 'w') as fd:
        fd.write(
            textwrap.dedent(
                f"""\
                SITE:
                    ID:  {site_id}
                """
            )
        )

    answer = Confirm.ask("Shall I add the environment to your ~/bash_profile ?")
    if answer:
        with open(Path("~/.bash_profile").expanduser(), 'a') as fd:
            fd.write(
                textwrap.dedent(
                    f"""
                    # Environment for project {project} added by `cgse init` at {format_datetime()}
                    export PROJECT={project}
                    export SITE_ID={site_id}
                    export {project}_DATA_STORAGE_LOCATION={data_storage_location}
                    export {project}_CONF_DATA_LOCATION={conf_data_location}
                    export {project}_LOG_FILE_LOCATION={log_file_location}
                    export {project}_LOCAL_SETTINGS={local_settings_path}
                    """
                )
            )
    else:
        rich.print(
            textwrap.dedent(
                f"""
                # -> Add the following lines to your bash profile or equivalent
                
                export PROJECT={project}
                export SITE_ID={site_id}
                export {project}_DATA_STORAGE_LOCATION={data_storage_location}
                export {project}_CONF_DATA_LOCATION={conf_data_location}
                export {project}_LOG_FILE_LOCATION={log_file_location}
                export {project}_LOCAL_SETTINGS={local_settings_path}
            """
            )

        )


show = typer.Typer(help="Show information about settings, environment, setup, ...", no_args_is_help=True)


@show.command(name="settings")
def show_settings():
    """Show the settings that are defined by the installed packages."""
    proc = subprocess.Popen(
        [sys.executable, "-m", "egse.settings"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    stdout, stderr = proc.communicate()
    rich.print(stdout.decode(), end='')
    if stderr:
        rich.print(f"[red]{stderr.decode()}[/]")


@show.command(name="env")
def show_env(
        mkdir: Annotated[bool, typer.Option(help="Create the missing folder")] = None,
        full: Annotated[bool, typer.Option(help="Provide additional info")] = None,
        doc: Annotated[bool, typer.Option(help="Provide documentation on environment variables")] = None,
):
    """Show the environment variables that are defined for the project."""
    options = [opt for opt, flag in [("--mkdir", mkdir), ("--full", full), ("--doc", doc)] if flag]

    cmd = [sys.executable, "-m", "egse.env"]
    cmd += options if options else []

    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    stdout, stderr = proc.communicate()
    rich.print(stdout.decode(), end='')
    if stderr:
        rich.print(f"[red]{stderr.decode()}[/]")


@show.command(name="procs")
def show_processes():
    """Show the processes that are running for the installed packages."""
    for ep in entry_points("cgse.explore"):
        # print(f"{ep.name = }, {ep.module = }, {ep.load() = }, {ep.extras = }")
        explore = ep.load()
        with contextlib.suppress(AttributeError):
            show_procs = getattr(explore, "show_processes")
            for line in show_procs():
                rich.print(line)


check = typer.Typer(help="Check installation, settings, required files, etc.", no_args_is_help=True)


@check.command(name="setups")
def check_setups():
    """Perform a number of checks on the SETUP files."""

    # What can we check with respect to the setups?
    #
    # - CONF_DATA_LOCATION

    from egse.env import get_conf_data_location
    from egse.env import get_site_id
    from egse.config import find_files

    any_errors = 0

    conf_data_location = get_conf_data_location()
    site_id = get_site_id()

    # ---------- check if the <PROJECT>_CONF_DATA_LOCATION is set

    if not conf_data_location:
        any_errors += 1
        rich.print("[red]The location of the configuration data can not be determined, check your environment.[/]")

    if not Path(conf_data_location).exists():
        any_errors += 1
        rich.print(f"[red]The location of the configuration data doesn't exist: {conf_data_location!s}[/]")

    # ---------- check if there is at least one SETUP in the configuration data folder

    files = list(find_files("SETUP*.yaml", root=conf_data_location))

    if not files:
        any_errors += 1
        rich.print(f"[red]No SETUP files were found at {conf_data_location}[/]")

    regex = re.compile(f"SETUP_{site_id}_00000_.*.yaml")

    if not any(True for file in files if regex.search(str(file))):
        any_errors += 1
        rich.print(f"[red]The is no Zero SETUP for {site_id} in {conf_data_location}[/]")

    if not any_errors:
        rich.print("[green]everything seems to be ok.[/]")



import shutil
from pathlib import Path

import pytest
from egse.resource import AmbiguityError
from egse.resource import NoSuchFileError
from egse.resource import ResourceError
from egse.resource import add_resource_id
from egse.resource import get_resource
from egse.resource import get_resource_dirs
from egse.resource import get_resource_locations
from egse.resource import get_resource_path
from egse.resource import initialise_resources
from egse.resource import print_resources
from egse.system import format_datetime
from fixtures.helpers import create_empty_file

HERE = Path(__file__).parent.resolve()


@pytest.fixture
def reset_resource():
    import egse.resource
    egse.resource.resources.clear()


def test_find_file_in_resource_location():

    print()

    data_root = Path(__file__).parent / "data"
    initialise_resources(data_root)
    print_resources()

    with pytest.raises(ResourceError):
        _ = get_resource(":/cars/volvo.txt")

    # Find file in first resource location, both with and without wildcard

    fn = get_resource(":/icons/sponge.svg")
    assert 'egse/gui/icons/sponge' in str(fn)

    fn = get_resource(":/icons/busy*.svg")
    assert 'egse/gui/icons/busy.svg' in str(fn)

    # Find file in second resource location, both with and without wildcard

    fn = get_resource(":/icons/keyboard.png")
    assert 'tests/data/icons/keyboard' in str(fn)

    with create_empty_file(filename=data_root / "icons" / f"file_with_timestamp_{format_datetime('today')}.txt"):
        fn = get_resource(":/icons/file_with_timestamp*.txt")
        assert f"tests/data/icons/file_with_timestamp_{format_datetime('today')}.txt" in str(fn)

    # File is not found in any resource location

    with pytest.raises(NoSuchFileError, match='unknown.png') as exc_info:
        _ = get_resource(":/icons/unknown.png")
    assert exc_info.type is NoSuchFileError
    assert "No file found that matches" in exc_info.value.args[0]


def test_find_file_in_resource_location_with_wildcard():

    print()

    data_root = Path(__file__).parent / "data"

    with create_empty_file(filename=data_root / "icons" / "flat" / "screen.png", create_folder=True):
        initialise_resources(data_root)
        print_resources()

        fn = get_resource(":/icons/*/screen.png")
        assert 'tests/data/icons/flat/screen' in str(fn)

        with create_empty_file(filename=data_root / "icons" / "flat" / f"file_with_timestamp_"
                                                                       f"{format_datetime('today')}.txt"):
            fn = get_resource(":/icons/*/file_with_timestamp*.txt")
            assert f"tests/data/icons/flat/file_with_timestamp_{format_datetime('today')}.txt" in str(fn)

    with create_empty_file(filename=data_root / "icons" / "double" / "flat" / "table.png", create_folder=True):
        initialise_resources(data_root)
        print_resources()

        fn = get_resource(":/icons/**/table.png")
        assert 'tests/data/icons/double/flat/table' in str(fn)

        with create_empty_file(
                filename=data_root / "icons" / "double" / "flat" / f"file_with_timestamp_{format_datetime()}.txt"
        ):
            fn = get_resource(":/icons/**/file_with_timestamp*.txt")
            assert "tests/data/icons/double/flat/file_with" in str(fn)


def test_initialisation(caplog):

    print()

    # The resource module initialises itself with root=__file__

    with pytest.raises(ResourceError):
        _ = get_resource(":/icons/some-image.png")

    # This will just issue a warning log message

    initialise_resources()
    print_resources()
    print(f"{caplog.text = }")

    caplog.clear()

    initialise_resources(Path(__file__).parent / "data")
    print_resources()
    print(f"{caplog.text = }")

    fn = get_resource(":/icons/sponge.svg")
    print(f"{fn = }")
    fn = get_resource(":/icons/keyboard.png")
    print(f"{fn = }")


def test_initialisation_ambiguity(reset_resource):

    print()

    # Let's first create a directory hierarchy that contains ambiguity in the sense that the same folder
    # exists at different levels in the hierarchy. `data` is a default resource location and we will use
    # that for this test.

    folders = (
            HERE / "x_resources" / "aaa" / "xdata",
            HERE / "x_resources" / "xxx" / "data",
            HERE / "x_resources" / "xxx" / "one" / "data",
            HERE / "x_resources" / "xxx" / "two" / "data",
            HERE / "x_resources" / "bbb" / "one" / "data",
    )

    for folder in folders:
        folder.mkdir(parents=True, exist_ok=True)

    # This function should not raise an AmbiguityError anymore

    initialise_resources(HERE / "x_resources")

    resource_locations = get_resource_locations()

    print(f"{resource_locations = }")

    for location in resource_locations['data']:
        if location.match("*/x_resources/bbb/one/data"):
            break
    else:
        pytest.fail("Expected resource not found!")

    shutil.rmtree(HERE / "x_resources")


def test_get_resource(reset_resource):

    print()

    # Need to call initialise here because we reset the module in the fixture

    initialise_resources(Path(__file__).parent / "data")

    pathname = get_resource(":/icons/keyboard.png")
    assert isinstance(pathname, Path)
    assert "board" in str(pathname)
    assert pathname.exists()

    pathname = get_resource(":/data/data-file.txt")
    assert isinstance(pathname, Path)
    assert "data-file" in str(pathname)
    assert pathname.exists()

    with pytest.raises(NoSuchFileError):
        _ = get_resource(":/icons/no-such-file.png")

    pathname = get_resource(__file__)
    print(f"{pathname = }")
    assert isinstance(pathname, Path)
    assert "resource" in str(pathname)
    assert pathname.exists()

    with pytest.raises(NoSuchFileError):
        _ = get_resource("resource.py")


def test_get_resource_with_wildcard(reset_resource):

    print()

    root = Path(__file__).parent / "data"
    initialise_resources(root)

    # Create test folder

    test_folder = root / "data/resources"
    test_folder.mkdir(parents=True, exist_ok=True)

    try:
        # Add test files

        for fn in "001_1000_RC2.txt", "002_1001_RC3.txt", "003_1001_RC3.txt", "004_1002_RC3.txt":
            create_empty_file(test_folder / fn)

        for fn in "005_2000_RC3.txt", "002_2001_RC4.txt", "003_2001_RC4.txt", "004_2002_RC4.txt":
            create_empty_file(test_folder / "two" / "levels" / fn, create_folder=True)

        add_resource_id("RCs", root / "data/resources")

        print_resources()

        rc_file = get_resource(":/RCs/001_*.txt")
        assert "RC2" in str(rc_file)
        assert "data/data/resources" in str(rc_file)

        rc_file = get_resource(":/RCs/005_*.txt")
        assert "RC3" in str(rc_file)
        assert "data/data/resources/two" in str(rc_file)

        with pytest.raises(AmbiguityError):
            _ = get_resource(":/RCs/*_1001_*.txt")

        with pytest.raises(AmbiguityError):
            _ = get_resource(":/RCs/*/*_2001_*.txt")

        _ = get_resource(":/RCs/*/002_2001_*.txt")
        _ = get_resource(":/RCs/**/002_2001_*.txt")

        with pytest.raises(AmbiguityError):
            _ = get_resource(":/RCs/**/00?_2001_*.txt")

        with pytest.raises(AmbiguityError):
            _ = get_resource(":/RCs/00?_1001_*.txt")

        with pytest.raises(AmbiguityError):
            _ = get_resource(":/RCs/00?_2001_*.txt")

        with pytest.raises(NoSuchFileError):
            _ = get_resource(":/RCs/00?_1005*.txt")

        with pytest.raises(NoSuchFileError):
            _ = get_resource(":/RCs/*/005_200?_RC4.txt")

        with pytest.raises(NoSuchFileError):
            _ = get_resource(":/RCs/**/007_200?_RC4.txt")

        with pytest.raises(ResourceError):
            _ = get_resource(":/abc/def.txt")

        with pytest.raises(NoSuchFileError):
            _ = get_resource(":/RCs/?/def.txt")

    finally:
        shutil.rmtree(test_folder)


def test_add_resource(reset_resource):

    print()

    assert get_resource_locations() == {}

    # Need to call initialise here because we reset the module in the fixture

    root = Path(__file__).parent / "data"
    initialise_resources(root)

    resources = get_resource_locations()

    assert 'icons' in resources
    assert 'data' in resources

    new_dir = "fancy-plugs"
    new_resource = "plugs"

    try:
        # ValueError because the folder `new_dirt` doesn't exist yet in 'root'

        with pytest.raises(ValueError):
            add_resource_id(new_resource, root / new_dir)

        (root / new_dir).mkdir()

        add_resource_id(new_resource, root / new_dir)

        n_styles = len(get_resource_locations().get("styles", []))

        (root / "xxx").mkdir()
        (root / "yyy").mkdir()

        add_resource_id("styles", root / "xxx")
        add_resource_id("styles", root / "yyy")

        resource_locations = get_resource_locations()
        assert "styles" in resource_locations
        assert len(resource_locations["styles"]) == n_styles + 2

    finally:
        # Remove folder created for this test

        for folder in root/new_dir, root/"xxx", root/"yyy":
            shutil.rmtree(folder)


def test_get_resource_dirs(caplog):

    assert len(get_resource_dirs(__file__)) > 0

    # We have a data folder in the tests, so check if this folder is part of the resource dirs

    assert any([str(x).endswith("data") for x in get_resource_dirs(__file__)])

    caplog.clear()

    assert get_resource_dirs(None) == []
    assert "WARNING" in caplog.text


def test_get_resource_path():

    with pytest.raises(FileNotFoundError):
        get_resource_path("non-existing-file")

    # The following resources should all exist in this project
    # We don't need the return value as the function raises a FileNotFoundError
    # when the resource is not found.

    _ = get_resource_path("empty_data_file.txt", Path(__file__).parent / "data")  # in tests/data/data
    _ = get_resource_path("keyboard.png", Path(__file__).parent / "data")  # in tests/data/icons



from pathlib import Path

import pytest
import rich

from egse.decorators import profile
from egse.env import env_var
from egse.settings import Settings
from egse.settings import SettingsError
from egse.settings import load_settings_file

HERE = Path(__file__).parent


def test_load_filename():
    # Specific test for a call with the filename and a location. This is the way the command files
    # will be loaded.

    settings = Settings.load(location=HERE / "data" / "data", filename="command.yaml")

    assert "Commands" in settings
    assert settings.Commands['list_commands']['description'] == "Returns a list of the available commands."


def test_load_settings_file():

    settings = load_settings_file(path=HERE / "data" / "data", filename="calibration.yaml", force=True)

    assert "cal_1" in settings
    assert settings.cal_1['coefficients'] ==  [1.5, 2.1, 3.4, 4.7]


def test_load_empty_settings_file():

    settings = load_settings_file(path=HERE / "data" / "data", filename="empty_yaml_file.yaml", force=True)

    assert settings == {}


def test_unknown_settings_file():

    with pytest.raises(SettingsError, match="Settings YAML file 'non-existing.yaml' is not found"):
        _ = load_settings_file(path=HERE / "data" / "data", filename="non-existing.yaml", force=True)


def test_corrupt_yaml_file():

    with pytest.raises(SettingsError, match="Error loading YAML document"):
        _ = load_settings_file(path=HERE / "data" / "data", filename="corrupt.yaml", force=True)


def test_load_global_settings():

    # Since this is a unit test for `cgse-common`, only that Settings YAML file should be loaded.

    settings = Settings.load()

    assert "SITE" in settings.keys()
    assert "PROCESS" in settings.keys()


def test_load_global_settings_group():

    settings = Settings.load("SITE")

    assert "ID" in settings.keys() and settings.ID == "XXXX"
    assert "SSH_SERVER" in settings.keys() and settings.SSH_SERVER == "localhost"
    assert "SSH_PORT" in settings.keys() and settings.SSH_PORT == 22

    with pytest.raises(SettingsError, match="Group name 'SITE_ID' is not defined"):
        _ = Settings.load("SITE_ID")


def test_load_local_settings():

    with env_var(CGSE_LOCAL_SETTINGS=str(HERE / "data" / "data" / "local_settings.yaml")):
        settings = Settings.load(add_local_settings=True)
        assert settings.SITE['ID'] == "YYYY"

        settings = Settings.load("SITE", add_local_settings=True)
        assert settings.ID == "YYYY"

        with pytest.raises(SettingsError, match="Group name 'SITE_ID' is not defined"):
            _ = Settings.load("SITE_ID", add_local_settings=True)


def test_get_site_id():

    from egse.settings import get_site_id

    Settings.clear_memoized()

    with env_var(CGSE_LOCAL_SETTINGS=None):
        site_id = get_site_id()
        assert site_id == "XXXX"  # This should be the SITE.ID from the settings.yaml file in cgse_common

    with env_var(CGSE_LOCAL_SETTINGS=str(HERE / "data" / "data" / "local_settings.yaml")):
        Settings.load("SITE", add_local_settings=True, force=True)
        site_id = get_site_id()
        assert site_id == "YYYY"  # This should be the SITE.ID from the local_settings.yaml file


def test_profiling(capsys):

    @profile
    def do_something():
        rich.print("Running do_something()")
        return Settings.load()

    x = do_something()

    captured = capsys.readouterr()
    assert "Running do_something()" in captured.out
    assert "PROFILE" not in captured.out
    assert captured.err == ""
    assert "PACKAGES" in x

    Settings.set_profiling(True)

    x = do_something()
    captured = capsys.readouterr()
    assert "Running do_something()" in captured.out
    assert "PROFILE[1]: " in captured.out
    assert captured.err == ""
    assert "PACKAGES" in x

    Settings.set_profiling(False)



"""
The Configuration Manager is a server that controls and distributes configuration settings.

The following functionality is provided:

* creation and distribution of observation identifiers
* start and end of observations or tests
* maintain proper Setups and distribute the latest Setup on demand

"""
import logging
import multiprocessing
import sys
from pathlib import Path
from typing import Annotated

import rich
import typer
import zmq
from prometheus_client import start_http_server

from egse.confman import ConfigurationManagerProtocol
from egse.confman import ConfigurationManagerProxy
from egse.control import ControlServer
from egse.env import get_conf_data_location
from egse.process import SubProcess
from egse.response import Failure
from egse.response import Response
from egse.settings import Settings
from egse.storage import store_housekeeping_information

# Use explicit name here otherwise the logger will probably be called __main__

logger = logging.getLogger("egse.confman.confman_cs")

CTRL_SETTINGS = Settings.load("Configuration Manager Control Server")


class ConfigurationManagerControlServer(ControlServer):
    def __init__(self):
        super().__init__()

        self.device_protocol = ConfigurationManagerProtocol(self)

        self.logger = logger
        self.logger.debug(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.set_hk_delay(10.0)

        self.logger.info(f"CM housekeeping saved every {self.hk_delay / 1000:.1f} seconds.")

    def get_communication_protocol(self):
        return CTRL_SETTINGS.PROTOCOL

    def get_commanding_port(self):
        return CTRL_SETTINGS.COMMANDING_PORT

    def get_service_port(self):
        return CTRL_SETTINGS.SERVICE_PORT

    def get_monitoring_port(self):
        return CTRL_SETTINGS.MONITORING_PORT

    def get_storage_mnemonic(self):
        try:
            return CTRL_SETTINGS.STORAGE_MNEMONIC
        except AttributeError:
            return "CM"

    def is_storage_manager_active(self):
        from egse.storage import is_storage_manager_active
        return is_storage_manager_active()

    def store_housekeeping_information(self, data):
        """Send housekeeping information to the Storage manager."""

        origin = self.get_storage_mnemonic()
        store_housekeeping_information(origin, data)

    def register_to_storage_manager(self):
        from egse.storage import register_to_storage_manager
        from egse.storage.persistence import TYPES

        register_to_storage_manager(
            origin=self.get_storage_mnemonic(),
            persistence_class=TYPES["CSV"],
            prep={
                "column_names": list(self.device_protocol.get_housekeeping().keys()),
                "mode": "a",
            }
        )

    def unregister_from_storage_manager(self):
        from egse.storage import unregister_from_storage_manager

        unregister_from_storage_manager(origin=self.get_storage_mnemonic())

    def before_serve(self):
        start_http_server(CTRL_SETTINGS.METRICS_PORT)


app = typer.Typer(name="cm_cs", no_args_is_help=True)


@app.command()
def start():
    """
    Starts the Configuration Manager (cm_cs). The cm_cs is a server which handles the
    configuration (aka Setup) of your test system.

    The cm_cs is normally started automatically on egse-server boot.
    """

    multiprocessing.current_process().name = "cm_cs"

    try:
        check_prerequisites()
    except RuntimeError as exc:
        logger.info(exc)
        return 0

    try:
        control_server = ConfigurationManagerControlServer()
        control_server.serve()
    except KeyboardInterrupt:
        print("Shutdown requested...exiting")
    except SystemExit as exit_code:
        print(f"System Exit with code {exit_code}.")
        sys.exit(exit_code.code)
    except Exception:
        import traceback

        traceback.print_exc(file=sys.stdout)

    return 0


@app.command()
def start_bg():
    """Start the Configuration Manager Control Server in the background."""
    proc = SubProcess("cm_cs", ["cm_cs", "start"])
    proc.execute()


@app.command()
def stop():
    """Send a 'quit_server' command to the Configuration Manager."""
    try:
        with ConfigurationManagerProxy() as cm:
            sp = cm.get_service_proxy()
            sp.quit_server()
    except ConnectionError:
        rich.print("[red]ERROR: Couldn't connect to the configuration manager.[/]")


@app.command()
def status():
    """Print the status of the control server."""

    import rich
    from egse.confman import get_status

    rich.print(get_status(), end='')


@app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def list_setups(ctx: Annotated[typer.Context, typer.Option()] = None):
    """List available Setups."""

    args = ctx.args if ctx else []

    # These extra arguments need to be converted into a dictionary as expected by
    # the function egse.system.filter_by_attr()
    for extra_arg in args:
        print(f"Got extra arg: {extra_arg}")

    with ConfigurationManagerProxy() as cm:
        setups = cm.list_setups(**{})
    if isinstance(setups, Failure):
        rich.print(f"[red]ERROR: {setups}[/]")
        rich.print("Check the log file for a more detailed error message.")
        return
    if setups:
        # We want to have the most recent (highest id number) last, but keep the site together
        setups = sorted(setups, key=lambda x: (x[1], x[0]))
        print("\n".join(f"{setup}" for setup in setups))


@app.command()
def load_setup(setup_id: int):
    """Load the given Setup on the configuration manager."""

    with ConfigurationManagerProxy() as cm:
        setup = cm.load_setup(setup_id)
    if isinstance(setup, Response):
        print(setup)
        return
    if setup.has_private_attribute("_setup_id"):
        setup_id = setup.get_private_attribute("_setup_id")
        print(f"{setup_id} loaded on configuration manager.")


@app.command()
def reload_setups():
    """ Clears the cache and re-loads the available setups.

    Note that this does not affect the currently loaded setup.
    """

    with ConfigurationManagerProxy() as pm:
        pm.reload_setups()


def check_prerequisites():
    """Checks if all prerequisites for running the Configuration Manager are met.

    Raises:
        RuntimeError when one or more of the prerequisites is not met.
    """

    fails = 0

    # We need a proper location for storing the configuration data.

    location = get_conf_data_location()

    if not location:
        raise RuntimeError(
            "The location for the configuration data is not defined. Please check your environment."
        )

    location = Path(location)

    if not location.exists():
        logger.error(
            f"The directory {location} does not exist, provide a writable location for "
            f"storing the configuration data."
        )
        fails += 1

    logger.debug(f"location = {location}")

    # now raise the final verdict

    if fails:
        raise RuntimeError(
            "Some of the prerequisites for the Configuration Manager haven't met. "
            "Please check the logs."
        )


if __name__ == "__main__":

    sys.exit(app())
